"""Tests the `google._utils.convert_message_params` function."""

import io
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from google.genai.types import ContentDict

from mirascope.core.base import (
    AudioPart,
    AudioURLPart,
    BaseMessageParam,
    CacheControlPart,
    DocumentPart,
    DocumentURLPart,
    ImagePart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.google._utils._convert_message_params import convert_message_params


@patch("PIL.Image.open", new_callable=MagicMock)
def test_convert_message_params(mock_image_open: MagicMock) -> None:
    """Tests the `convert_message_params` function."""
    mock_image_open.return_value = "test"
    message_params: list[BaseMessageParam | ContentDict] = [
        BaseMessageParam(role="system", content="You are a helpful assistant."),
        BaseMessageParam(role="user", content="Hello"),
        {"role": "user", "parts": ["Hello", {"type": "text", "text": "Hello"}]},  # pyright: ignore[reportAssignmentType]
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="test"),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail=None
                ),
                AudioPart(type="audio", media_type="audio/wav", audio=b"audio"),
                DocumentPart(
                    type="document", media_type="application/pdf", document=b"%PDF-1.4"
                ),
            ],
        ),
    ]
    mock_client = MagicMock()
    converted_message_params = convert_message_params(message_params, mock_client)
    assert converted_message_params == [
        {"parts": [{"text": "You are a helpful assistant."}], "role": "system"},
        {"parts": [{"text": "Hello"}], "role": "user"},
        {"parts": ["Hello", {"text": "Hello", "type": "text"}], "role": "user"},
        {
            "parts": [
                {"text": "test"},
                {"inline_data": {"data": b"image", "mime_type": "image/jpeg"}},
                {"inline_data": {"data": b"audio", "mime_type": "audio/wav"}},
                {"inline_data": {"data": b"%PDF-1.4", "mime_type": "application/pdf"}},
            ],
            "role": "user",
        },
    ]

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. Google currently only supports "
        "JPEG, PNG, WebP, HEIC, and HEIF images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail=None,
                        )
                    ],
                ),
            ],
            mock_client,
        )

    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/unknown. "
        "Google currently only supports WAV, MP3, AIFF, AAC, OGG, "
        "and FLAC audio file types.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(
                            type="audio", media_type="audio/unknown", audio=b"audio"
                        )
                    ],
                ),
            ],
            mock_client,
        )

    with pytest.raises(
        ValueError,
        match="Google currently only supports text, tool_call, tool_result, image, audio, and document parts. Part provided: cache_control",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        CacheControlPart(type="cache_control", cache_type="ephemeral")
                    ],
                )
            ],
            mock_client,
        )

    with pytest.raises(
        ValueError,
        match="Unsupported document media type: application/msword. Google currently only supports PDF, JavaScript, Python, TXT, HTML, CSS, CSV, XML, RTF, and Markdown document types.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        DocumentPart(
                            type="document",
                            document=b"doc_data",
                            media_type="application/msword",
                        )
                    ],
                )
            ],
            mock_client,
        )


def test_document_url_with_http_valid() -> None:
    """Test document_url part with an HTTP URL and valid media type."""
    message = BaseMessageParam(
        role="user",
        content=[
            DocumentURLPart(type="document_url", url="https://example.com/document.pdf")
        ],
    )
    mock_client = MagicMock()
    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "role": "user",
            "parts": [
                {
                    "file_data": {
                        "file_uri": "https://example.com/document.pdf",
                        "mime_type": None,
                    }
                }
            ],
        }
    ]


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"%PDF-1.4...",
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_document_type",
    return_value="pdf",
)
def test_document_url_with_http_valid_gemini(
    mock_get_document_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test document_url part with an HTTP URL and valid media type."""
    message = BaseMessageParam(
        role="user",
        content=[
            DocumentURLPart(type="document_url", url="https://example.com/document.pdf")
        ],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_file = MagicMock(
        uri="https://example.com/document.pdf", mime_type="application/pdf"
    )
    mock_client.files.upload.return_value = mock_file
    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "role": "user",
            "parts": [
                {
                    "inline_data": {
                        "data": b"%PDF-1.4...",
                        "mime_type": "application/pdf",
                    }
                }
            ],
        }
    ]
    mock_load_media.assert_called_once_with("https://example.com/document.pdf")
    mock_get_document_type.assert_called_once_with(b"%PDF-1.4...")


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"\x80\x81\x83",
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_document_type",
    return_value="unknown",
)
def test_document_url_invalid(
    mock_get_document_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test url with an invalid media type raises ValueError."""
    message = BaseMessageParam(
        role="user",
        content=[
            DocumentURLPart(type="document_url", url="https://example.com/document.exe")
        ],
    )

    expected_error_message = (
        "Failed to process document from URL: https://example.com/document.exe. "
        "Error details: Unsupported document format detected for URL: https://example.com/document.exe. "
        "Google API only supports the following document formats: "
        "PDF, JavaScript, Python, TXT, HTML, CSS, Markdown, CSV, XML and RTF. "
        "Please provide a document in one of these supported formats.. "
        "Please ensure the URL is accessible and points to a supported document format."
    )
    with pytest.raises(
        ValueError,
        match=expected_error_message,
    ):
        mock_client = MagicMock()
        mock_client.vertexai = False
        convert_message_params([message], mock_client)

    mock_load_media.assert_called_once_with("https://example.com/document.exe")
    mock_get_document_type.assert_called_once_with(b"\x80\x81\x83")


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
)
def test_document_url_non_http(mock_load_media: MagicMock) -> None:
    """Test document_url part with a non-HTTP URL returns a FileData object."""
    message = BaseMessageParam(
        role="user",
        content=[DocumentURLPart(type="document_url", url="gs://bucket/document.pdf")],
    )

    result = convert_message_params([message], MagicMock())

    assert len(result) == 1
    assert "parts" in result[0]
    assert result[0]["parts"]
    part = result[0]["parts"][0]
    assert isinstance(part, dict)
    assert part == {
        "file_data": {
            "file_uri": "gs://bucket/document.pdf",
            "mime_type": None,
        }
    }

    mock_load_media.assert_not_called()


def test_image_url_with_http_valid() -> None:
    """Test image_url part with an HTTP URL and valid media type."""
    message = BaseMessageParam(
        role="user",
        content=[
            ImageURLPart(type="image_url", url="https://example.com/image", detail=None)
        ],
    )
    result = convert_message_params([message], MagicMock())
    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "https://example.com/image",
                        "mime_type": None,
                    }
                }
            ],
            "role": "user",
        }
    ]


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"image_url_data",
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_image_type",
    return_value="png",
)
def test_image_url_with_http_valid_gemini(
    mock_get_image_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test image_url part with an HTTP URL and valid media type."""
    message = BaseMessageParam(
        role="user",
        content=[
            ImageURLPart(type="image_url", url="https://example.com/image", detail=None)
        ],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_file = MagicMock(uri="file://local/path/image", mime_type="image/png")
    mock_client.files.upload.return_value = mock_file
    result = convert_message_params([message], mock_client)
    assert result == [
        {
            "parts": [
                {
                    "inline_data": {
                        "data": b"image_url_data",
                        "mime_type": "image/png",
                    }
                }
            ],
            "role": "user",
        }
    ]
    # Check that _load_media was called with the URL.
    mock_load_media.assert_called_once_with("https://example.com/image")


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"image_url_data",
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_image_type",
    return_value="svg",
)
def test_image_url_with_http_invalid_media_type(
    mock_get_image_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test image_url part with an HTTP URL and an invalid media type raises ValueError."""
    message = BaseMessageParam(
        role="user",
        content=[
            ImageURLPart(type="image_url", url="https://example.com/image", detail=None)
        ],
    )
    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. Google currently only supports JPEG, PNG, WebP, HEIC, and HEIF images.",
    ):
        client = MagicMock()
        client.vertexai = False
        convert_message_params([message], client)


@patch("mirascope.core.google._utils._convert_message_params._load_media")
def test_image_url_with_non_http(mock_load_media: MagicMock) -> None:
    """Test image_url part with a non-HTTP URL returns a FileData object."""
    # For non-HTTP URLs, _load_media should not be called.
    message = BaseMessageParam(
        role="user",
        content=[
            ImageURLPart(type="image_url", url="file://local/path/image", detail=None)
        ],
    )
    result = convert_message_params([message], MagicMock())
    # Expect the part to be a FileData with file_uri equal to the URL.
    assert len(result) == 1
    assert "parts" in result[0]
    assert result[0]["parts"]
    part = result[0]["parts"][0]
    assert isinstance(part, dict)
    assert part == {
        "file_data": {
            "file_uri": "file://local/path/image",
            "mime_type": None,
        }
    }
    mock_load_media.assert_not_called()


def test_audio_url_with_http_valid() -> None:
    """Test audio_url part with an HTTP URL and a valid audio type."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="https://example.com/audio")],
    )
    result = convert_message_params([message], MagicMock())
    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "https://example.com/audio",
                        "mime_type": None,
                    }
                },
            ],
            "role": "user",
        }
    ]


@patch(
    "mirascope.core.google._utils._convert_message_params.get_audio_type",
    return_value="wav",
)
@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"audio_data",
)
def test_audio_url_with_http_valid_gemini(
    mock_load_media: MagicMock, mock_get_audio_type: MagicMock
) -> None:
    """Test image_url part with an HTTP URL and valid media type."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="https://example.com/audio")],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_file = MagicMock(uri="file://local/path/audio", mime_type="audio/wav")
    mock_client.files.upload.return_value = mock_file
    result = convert_message_params([message], mock_client)
    assert result == [
        {
            "parts": [
                {
                    "inline_data": {
                        "data": b"audio_data",
                        "mime_type": "audio/wav",
                    }
                }
            ],
            "role": "user",
        }
    ]
    mock_load_media.assert_called_once_with("https://example.com/audio")


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"audio_data",
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_audio_type",
    return_value="unknown",
)
def test_audio_url_with_http_invalid(
    mock_get_audio_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test audio_url part with an HTTP URL and an invalid audio type raises ValueError."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="https://example.com/audio")],
    )
    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/unknown. Google currently only supports WAV, MP3, AIFF, AAC, OGG, and FLAC audio file types.",
    ):
        client = MagicMock()
        client.vertexai = False
        convert_message_params([message], client)
    mock_load_media.assert_called_once_with("https://example.com/audio")
    mock_get_audio_type.assert_called_once_with(b"audio_data")


def test_audio_url_with_non_http() -> None:
    """Test audio_url part with a non-HTTP URL returns a FileData object."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="ftp://example.com/audio")],
    )
    result = convert_message_params([message], MagicMock())
    assert len(result) == 1
    assert "parts" in result[0]
    assert result[0]["parts"]
    part = result[0]["parts"][0]
    assert isinstance(part, dict)
    assert part == {
        "file_data": {
            "file_uri": "ftp://example.com/audio",
            "mime_type": None,
        }
    }


def test_async_large_image_upload() -> None:
    """Test async upload of large images (>15MB)."""
    # Create a large image (>15MB)
    large_image = bytes([0] * (22 * 1024 * 1024))  # 22MB

    message = BaseMessageParam(
        role="user",
        content=[
            ImagePart(
                type="image", media_type="image/jpeg", image=large_image, detail=None
            )
        ],
    )

    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_file = MagicMock(uri="file://uploaded/image", mime_type="image/jpeg")
    mock_client.aio.files.upload = AsyncMock(return_value=mock_file)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "file://uploaded/image",
                        "mime_type": "image/jpeg",
                    }
                }
            ],
            "role": "user",
        }
    ]

    # Verify async upload was called
    mock_client.aio.files.upload.assert_called_once()
    call_args = mock_client.aio.files.upload.call_args
    assert isinstance(call_args[1]["file"], io.BytesIO)
    assert call_args[1]["config"]["mime_type"] == "image/jpeg"


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=bytes([0] * (22 * 1024 * 1024)),
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_image_type",
    return_value="jpeg",
)
def test_large_image_url_upload(
    mock_get_image_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test large image URL upload."""
    message = BaseMessageParam(
        role="user",
        content=[
            ImageURLPart(type="image_url", url="https://example.com/large", detail=None)
        ],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_file = MagicMock(uri="file://uploaded/image", mime_type="image/jpeg")
    mock_client.aio.files.upload = AsyncMock(return_value=mock_file)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "file://uploaded/image",
                        "mime_type": "image/jpeg",
                    }
                }
            ],
            "role": "user",
        }
    ]

    # Verify async upload was called
    mock_client.aio.files.upload.assert_called_once()
    call_args = mock_client.aio.files.upload.call_args
    assert isinstance(call_args[1]["file"], io.BytesIO)
    assert call_args[1]["config"]["mime_type"] == "image/jpeg"


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=bytes([0] * (22 * 1024 * 1024)),
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_audio_type",
    return_value="wav",
)
def test_large_audio_url_upload(
    mock_get_audio_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test large audio URL upload."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="https://example.com/large")],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_file = MagicMock(uri="file://uploaded/audio", mime_type="audio/wav")
    mock_client.aio.files.upload = AsyncMock(return_value=mock_file)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "file://uploaded/audio",
                        "mime_type": "audio/wav",
                    }
                }
            ],
            "role": "user",
        }
    ]

    # Verify async upload was called
    mock_client.aio.files.upload.assert_called_once()
    call_args = mock_client.aio.files.upload.call_args
    assert isinstance(call_args[1]["file"], io.BytesIO)
    assert call_args[1]["config"]["mime_type"] == "audio/wav"


def test_multiple_async_image_uploads() -> None:
    """Test multiple concurrent async image uploads."""
    large_image = bytes([0] * (16 * 1024 * 1024))  # 16MB

    message = BaseMessageParam(
        role="user",
        content=[
            ImagePart(
                type="image", media_type="image/jpeg", image=large_image, detail=None
            ),
            TextPart(type="text", text="Some text in between"),
            ImagePart(
                type="image", media_type="image/png", image=large_image, detail=None
            ),
        ],
    )

    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()

    mock_files = [
        MagicMock(uri=f"file://uploaded/image{i}", mime_type=f"image/{fmt}")
        for i, fmt in enumerate(["jpeg", "png"])
    ]
    mock_client.aio.files.upload = AsyncMock(side_effect=mock_files)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "file://uploaded/image0",
                        "mime_type": "image/jpeg",
                    }
                },
                {"text": "Some text in between"},
                {
                    "file_data": {
                        "file_uri": "file://uploaded/image1",
                        "mime_type": "image/png",
                    }
                },
            ],
            "role": "user",
        }
    ]

    assert mock_client.aio.files.upload.call_count == 2
    calls = mock_client.aio.files.upload.call_args_list
    assert calls[0][1]["config"]["mime_type"] == "image/jpeg"
    assert calls[1][1]["config"]["mime_type"] == "image/png"


def test_async_upload_error_handling() -> None:
    """Test error handling during async uploads."""
    large_image = bytes([0] * (16 * 1024 * 1024))  # 16MB

    message = BaseMessageParam(
        role="user",
        content=[
            ImagePart(
                type="image", media_type="image/jpeg", image=large_image, detail=None
            ),
        ],
    )

    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_client.aio.files.upload = AsyncMock(side_effect=Exception("Upload failed"))

    with pytest.raises(Exception) as exc_info:
        convert_message_params([message], mock_client)
    assert str(exc_info.value) == "Upload failed"


def test_mixed_content_with_async_uploads() -> None:
    """Test handling mixed content types with some requiring async uploads. And also testing that only images
    actually put in the message are counted towards the 15MB limit."""
    large_image = bytes([0] * (16 * 1024 * 1024))  # 16MB
    small_image = bytes([0] * (1 * 1024 * 1024))  # 1MB

    message = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Start text"),
            ImagePart(
                type="image", media_type="image/jpeg", image=large_image, detail=None
            ),
            ImagePart(
                type="image", media_type="image/png", image=small_image, detail=None
            ),
            TextPart(type="text", text="End text"),
        ],
    )

    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_file = MagicMock(uri="file://uploaded/large_image", mime_type="image/jpeg")
    mock_client.aio.files.upload = AsyncMock(return_value=mock_file)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {"text": "Start text"},
                {
                    "file_data": {
                        "file_uri": "file://uploaded/large_image",
                        "mime_type": "image/jpeg",
                    }
                },
                {
                    "inline_data": {
                        "data": small_image,
                        "mime_type": "image/png",
                    }
                },
                {"text": "End text"},
            ],
            "role": "user",
        }
    ]

    # Verify only large image was uploaded async
    mock_client.aio.files.upload.assert_called_once()
    call_args = mock_client.aio.files.upload.call_args
    assert call_args[1]["config"]["mime_type"] == "image/jpeg"


def test_audio_part_handling() -> None:
    """Test handling of audio parts including large and small audio files."""
    large_audio = bytes([0] * (16 * 1024 * 1024))  # 16MB
    small_audio = bytes([0] * (1 * 1024 * 1024))  # 1MB

    message = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="Start text"),
            AudioPart(type="audio", media_type="audio/wav", audio=large_audio),
            AudioPart(type="audio", media_type="audio/mp3", audio=small_audio),
            TextPart(type="text", text="End text"),
        ],
    )

    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_file = MagicMock(uri="file://uploaded/large_audio", mime_type="audio/wav")
    mock_client.aio.files.upload = AsyncMock(return_value=mock_file)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {"text": "Start text"},
                {
                    "file_data": {
                        "file_uri": "file://uploaded/large_audio",
                        "mime_type": "audio/wav",
                    }
                },
                {
                    "inline_data": {
                        "data": small_audio,
                        "mime_type": "audio/mp3",
                    }
                },
                {"text": "End text"},
            ],
            "role": "user",
        }
    ]

    # Verify only large audio was uploaded async
    mock_client.aio.files.upload.assert_called_once()
    call_args = mock_client.aio.files.upload.call_args
    assert call_args[1]["config"]["mime_type"] == "audio/wav"


@pytest.mark.asyncio
async def test_convert_message_in_running_loop():
    """Tests convert_message_params in an async context."""
    result = convert_message_params([], MagicMock())
    assert result == []


def test_tool_call_parts() -> None:
    """Test handling of tool_call parts."""
    message = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="I need to search for something"),
            ToolCallPart(
                type="tool_call",
                name="search",
                args={"query": "python programming"},
                id="tool-1234",
            ),
        ],
    )

    result = convert_message_params([message], MagicMock())

    assert result == [
        {
            "parts": [
                {"text": "I need to search for something"},
                {
                    "function_call": {
                        "name": "search",
                        "args": {"query": "python programming"},
                        "id": "tool-1234",
                    }
                },
            ],
            "role": "user",
        }
    ]


def test_tool_result_parts() -> None:
    """Test handling of tool_result parts."""
    message = BaseMessageParam(
        role="assistant",
        content=[
            TextPart(type="text", text="Here are the search results:"),
            ToolResultPart(
                type="tool_result",
                name="search",
                content="Python is a programming language",
                id="tool-1234",
            ),
        ],
    )

    result = convert_message_params([message], MagicMock())

    assert result == [
        {
            "parts": [
                {"text": "Here are the search results:"},
                {
                    "function_response": {
                        "name": "search",
                        "response": {
                            "content": "Python is a programming language",
                        },
                        "id": "tool-1234",
                    }
                },
            ],
            "role": "model",
        }
    ]


def test_multiple_async_document_uploads() -> None:
    """Test multiple concurrent async document uploads."""
    large_document_data = bytes([0] * (25 * 1024 * 1024))

    message = BaseMessageParam(
        role="user",
        content=[
            DocumentPart(
                type="document",
                media_type="application/pdf",
                document=large_document_data,
            ),
            TextPart(type="text", text="Some text in between"),
            DocumentPart(
                type="document",
                media_type="application/pdf",
                document=large_document_data,
            ),
        ],
    )

    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()

    mock_files = [
        MagicMock(uri=f"file://uploaded/document{i}", mime_type="application/pdf")
        for i in range(2)
    ]
    mock_client.aio.files.upload = AsyncMock(side_effect=mock_files)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "file://uploaded/document0",
                        "mime_type": "application/pdf",
                    }
                },
                {"text": "Some text in between"},
                {
                    "file_data": {
                        "file_uri": "file://uploaded/document1",
                        "mime_type": "application/pdf",
                    }
                },
            ],
            "role": "user",
        }
    ]

    assert mock_client.aio.files.upload.call_count == 2
    calls = mock_client.aio.files.upload.call_args_list
    assert calls[0][1]["config"]["mime_type"] == "application/pdf"
    assert calls[1][1]["config"]["mime_type"] == "application/pdf"


def test_large_document_upload() -> None:
    """Test async upload of large documents (>10MB)."""
    large_document_data = bytes([0] * (25 * 1024 * 1024))  # 25MB

    message = BaseMessageParam(
        role="user",
        content=[
            DocumentPart(
                type="document",
                media_type="application/pdf",
                document=large_document_data,
            )
        ],
    )

    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_file = MagicMock(uri="file://uploaded/document", mime_type="application/pdf")
    mock_client.aio.files.upload = AsyncMock(return_value=mock_file)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "file://uploaded/document",
                        "mime_type": "application/pdf",
                    }
                }
            ],
            "role": "user",
        }
    ]

    mock_client.aio.files.upload.assert_called_once()
    call_args = mock_client.aio.files.upload.call_args
    assert isinstance(call_args[1]["file"], io.BytesIO)
    assert call_args[1]["file"].getvalue() == large_document_data
    assert call_args[1]["config"]["mime_type"] == "application/pdf"


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=bytes([0] * (25 * 1024 * 1024)),
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_document_type",
    return_value="pdf",
)
def test_large_document_url_upload(
    mock_get_document_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test large document URL upload."""
    message = BaseMessageParam(
        role="user",
        content=[
            DocumentURLPart(
                type="document_url", url="https://example.com/large_document.pdf"
            )
        ],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False
    mock_client.aio = MagicMock()
    mock_client.aio.files = MagicMock()
    mock_file = MagicMock(
        uri="file://uploaded/large_document", mime_type="application/pdf"
    )
    mock_client.aio.files.upload = AsyncMock(return_value=mock_file)

    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "file_data": {
                        "file_uri": "file://uploaded/large_document",
                        "mime_type": "application/pdf",
                    }
                }
            ],
            "role": "user",
        }
    ]

    mock_load_media.assert_called_once_with("https://example.com/large_document.pdf")
    mock_get_document_type.assert_called_once_with(bytes([0] * (25 * 1024 * 1024)))

    mock_client.aio.files.upload.assert_called_once()
    call_args = mock_client.aio.files.upload.call_args
    assert isinstance(call_args[1]["file"], io.BytesIO)
    assert call_args[1]["file"].getvalue() == bytes([0] * (25 * 1024 * 1024))
    assert call_args[1]["config"]["mime_type"] == "application/pdf"


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"This is a plain text document.",
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_document_type",
    side_effect=ValueError("Unsupported document type"),
)
def test_document_url_infers_text_plain(
    mock_get_document_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Tests that a document from a URL is inferred as text/plain if decodable."""
    message = BaseMessageParam(
        role="user",
        content=[
            DocumentURLPart(type="document_url", url="https://example.com/document.txt")
        ],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False
    result = convert_message_params([message], mock_client)

    assert result == [
        {
            "parts": [
                {
                    "inline_data": {
                        "data": b"This is a plain text document.",
                        "mime_type": "text/plain",
                    }
                }
            ],
            "role": "user",
        }
    ]
    mock_load_media.assert_called_once_with("https://example.com/document.txt")
    mock_get_document_type.assert_called_once_with(b"This is a plain text document.")


@patch(
    "mirascope.core.google._utils._convert_message_params._load_media",
    return_value=b"\xff\xfe\xfd",
)
@patch(
    "mirascope.core.google._utils._convert_message_params.get_document_type",
    side_effect=ValueError("Unknown type from get_document_type"),
)
def test_document_url_unsupported_type_after_decode_fails(
    mock_get_document_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Tests that an error is raised for unsupported doc type if UTF-8 decode fails."""
    message = BaseMessageParam(
        role="user",
        content=[
            DocumentURLPart(
                type="document_url", url="https://example.com/unknown_document"
            )
        ],
    )
    mock_client = MagicMock()
    mock_client.vertexai = False

    expected_error_message = (
        "Failed to process document from URL: https://example.com/unknown_document. "
        "Error details: Unsupported document format detected for URL: https://example.com/unknown_document. "
        "Google API only supports the following document formats: "
        "PDF, JavaScript, Python, TXT, HTML, CSS, Markdown, CSV, XML and RTF. "
        "Please provide a document in one of these supported formats.. "
        "Please ensure the URL is accessible and points to a supported document format."
    )
    with pytest.raises(ValueError, match=expected_error_message):
        convert_message_params([message], mock_client)

    mock_load_media.assert_called_once_with("https://example.com/unknown_document")
    mock_get_document_type.assert_called_once_with(b"\xff\xfe\xfd")



# file: gemini_message_param_converter.py

from typing import cast

from google.generativeai.types import (
    ContentDict,
    protos,
)

from mirascope.core import BaseMessageParam
from mirascope.core.base import (
    AudioPart,
    AudioURLPart,
    DocumentPart,
    ImagePart,
    ImageURLPart,
    TextPart,
)
from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)
from mirascope.core.base.message_param import ToolCallPart, ToolResultPart
from mirascope.core.gemini._utils import convert_message_params


def _is_image_mime(mime_type: str) -> bool:
    return mime_type in ["image/jpeg", "image/png", "image/gif", "image/webp"]


def _to_image_part(mime_type: str, data: bytes) -> ImagePart:
    if not _is_image_mime(mime_type):
        raise ValueError(
            f"Unsupported image media type: {mime_type}. "
            "Expected one of: image/jpeg, image/png, image/gif, image/webp."
        )
    return ImagePart(type="image", media_type=mime_type, image=data, detail=None)


def _is_audio_mime(mime_type: str) -> bool:
    return mime_type in [
        "audio/wav",
        "audio/mp3",
        "audio/wav",
        "audio/mp3",
        "audio/aiff",
        "audio/aac",
        "audio/ogg",
        "audio/flac",
    ]


def _to_audio_part(mime_type: str, data: bytes) -> AudioPart:
    if not _is_audio_mime(mime_type):
        raise ValueError(
            f"Unsupported audio media type: {mime_type}. "
            "Expected one of: audio/wav, audio/mp3, audio/aiff, audio/aac, audio/ogg, audio/flac."
        )
    return AudioPart(type="audio", media_type=mime_type, audio=data)


def _to_document_part(mime_type: str, data: bytes) -> DocumentPart:
    if mime_type != "application/pdf":
        raise ValueError(
            f"Unsupported document media type: {mime_type}. "
            "Only application/pdf is supported."
        )
    return DocumentPart(type="document", media_type=mime_type, document=data)


class GeminiMessageParamConverter(BaseMessageParamConverter):
    """Converts between Gemini `ContentDict` and Mirascope `BaseMessageParam`."""

    @staticmethod
    def to_provider(message_params: list[BaseMessageParam]) -> list[ContentDict]:
        """
        Convert from Mirascope `BaseMessageParam` to Gemini `ContentDict`.
        """
        return convert_message_params(
            cast(list[BaseMessageParam | ContentDict], message_params)
        )

    @staticmethod
    def from_provider(message_params: list[ContentDict]) -> list[BaseMessageParam]:
        """
        Convert from Gemini's `ContentDict` to Mirascope `BaseMessageParam`.
        """
        converted: list[BaseMessageParam] = []
        for message_param in message_params:
            role: str = (
                "assistant"
                if message_param["role"] == "model"
                else message_param["role"]
            )
            content_list = []
            for part in cast(
                list[protos.Part | protos.FunctionCall | protos.FunctionResponse],
                message_param["parts"],
            ):
                if isinstance(part, protos.Part):
                    if part.text:
                        content_list.append(TextPart(type="text", text=part.text))

                    elif part.inline_data:
                        blob = part.inline_data
                        mime = blob.mime_type
                        data = blob.data
                        if _is_image_mime(mime):
                            content_list.append(_to_image_part(mime, data))
                        elif _is_audio_mime(mime):
                            content_list.append(_to_audio_part(mime, data))
                        elif mime == "application/pdf":
                            content_list.append(_to_document_part(mime, data))
                        else:
                            raise ValueError(
                                f"Unsupported inline_data mime type: {mime}. Cannot convert to BaseMessageParam."
                            )

                    elif part.file_data:
                        if _is_image_mime(part.file_data.mime_type):
                            content_list.append(
                                ImageURLPart(
                                    type="image_url",
                                    url=part.file_data.file_uri,
                                    detail=None,
                                )
                            )
                        elif _is_audio_mime(part.file_data.mime_type):
                            content_list.append(
                                AudioURLPart(
                                    type="audio_url",
                                    url=part.file_data.file_uri,
                                )
                            )
                        else:
                            raise ValueError(
                                f"Unsupported file_data mime type: {part.file_data.mime_type}. Cannot convert to BaseMessageParam."
                            )
                    elif part.function_call:
                        converted.append(
                            BaseMessageParam(
                                role=role,
                                content=[
                                    ToolCallPart(
                                        type="tool_call",
                                        name=part.function_call.name,
                                        args=dict(part.function_call.args),
                                    )
                                ],
                            )
                        )
                    elif part.function_response:
                        converted.append(
                            BaseMessageParam(
                                role=role,
                                content=[
                                    ToolResultPart(
                                        type="tool_result",
                                        name=part.function_response.name,
                                        content=part.function_response.response[  # pyright: ignore [reportArgumentType]
                                            "result"
                                        ],
                                        id=None,
                                        is_error=False,
                                    )
                                ],
                            )
                        )
                    else:
                        raise ValueError(
                            "Part does not contain any supported content (text, image, or document)."
                        )
                elif isinstance(part, protos.FunctionCall):
                    converted.append(
                        BaseMessageParam(
                            role=role,
                            content=[
                                ToolCallPart(
                                    type="tool_call",
                                    name=part.name,
                                    args=dict(part.args),
                                )
                            ],
                        )
                    )
                elif isinstance(part, protos.FunctionResponse):
                    converted.append(
                        BaseMessageParam(
                            role=role,
                            content=[
                                ToolResultPart(
                                    type="tool_result",
                                    name=part.name,
                                    content=part.response["result"],  # pyright: ignore [reportArgumentType]
                                    id=None,
                                    is_error=False,
                                )
                            ],
                        )
                    )

            if len(content_list) == 1 and isinstance(content_list[0], TextPart):
                converted.append(
                    BaseMessageParam(role=role, content=content_list[0].text)
                )
            else:
                if content_list:
                    converted.append(BaseMessageParam(role=role, content=content_list))

        return converted



from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from pydantic import BaseModel

from mirascope.core.base import BaseCallResponse
from mirascope.core.base.stream import (
    BaseStream,
)
from mirascope.core.base.structured_stream import BaseStructuredStream
from mirascope.core.base.types import CostMetadata
from mirascope.core.bedrock import AssistantMessageTypeDef
from mirascope.integrations._middleware_factory import (
    default_context_manager,
    middleware_factory,
)


def test_default_context_manager() -> None:
    with default_context_manager(lambda x: x) as result:
        assert result is None


def test_default_context_manager_async() -> None:
    async def async_fn(x):
        return x  # pragma: no cover

    with default_context_manager(async_fn) as result:
        assert result is None


def test_middleware_factory_call_response_sync() -> None:
    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore

    def sync_fn() -> MyCallResponse:
        return call_response

    def handle_call_response(result, fn, context) -> None:
        assert isinstance(result, BaseCallResponse)

    decorate = middleware_factory(handle_call_response=handle_call_response)(sync_fn)
    result = decorate()
    assert result.content == call_response.content


@pytest.mark.asyncio
async def test_middleware_factory_call_response_async() -> None:
    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore

    async def async_fn() -> MyCallResponse:
        return call_response

    async def handle_call_response_async(result, fn, context) -> None:
        assert isinstance(result, BaseCallResponse)

    decorate = middleware_factory(
        handle_call_response_async=handle_call_response_async
    )(async_fn)
    result = await decorate()
    assert result.content == call_response.content


def test_middleware_factory_stream_sync() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    mock_chunk = MagicMock()
    mock_chunk.content = "content"
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    class MyStream(BaseStream):
        @property
        def cost(self) -> int:
            return 10

    # Usage
    stream_chunks = [(mock_chunk, None)]
    my_stream = MyStream(
        stream=(t for t in stream_chunks),
        metadata={},
        tool_types=[],
        call_response_type=MagicMock,
        model="model",
        prompt_template="prompt_template",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
    )  # type: ignore

    def sync_fn() -> BaseStream:
        return my_stream

    def handle_stream(result, fn, context) -> None:
        assert isinstance(result, BaseStream)

    decorate = middleware_factory(handle_stream=handle_stream)(sync_fn)
    result = decorate()
    result_chunks = []
    for chunk, _ in result:
        result_chunks.append(chunk.content)
    for result_chunk, stream_chunk in zip(result_chunks, stream_chunks, strict=False):
        assert result_chunk == stream_chunk[0].content
    assert my_stream.cost == result.cost


@pytest.mark.asyncio
async def test_middleware_factory_stream_async() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    mock_chunk = MagicMock()
    mock_chunk.content = "content"
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    class MyStream(BaseStream):
        @property
        def cost(self) -> int:
            return 10

    # Usage
    stream_chunks = [(mock_chunk, None)]
    my_stream = MyStream(
        stream=(t for t in stream_chunks),
        metadata={},
        tool_types=[],
        call_response_type=MagicMock,
        model="model",
        prompt_template="prompt_template",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
    )  # type: ignore

    async def generator():
        yield mock_chunk, None

    my_stream.stream = generator()

    async def async_fn() -> BaseStream:
        return my_stream

    async def handle_stream_async(result, fn, context) -> None:
        assert isinstance(result, BaseStream)

    decorate = middleware_factory(handle_stream_async=handle_stream_async)(async_fn)
    result = await decorate()
    result_chunks = []
    async for chunk, _ in result:
        result_chunks.append(chunk.content)
    for result_chunk, stream_chunk in zip(result_chunks, stream_chunks, strict=False):
        assert result_chunk == stream_chunk[0].content
    assert my_stream.cost == result.cost
    assert my_stream.cost_metadata == result.cost_metadata


def test_middleware_factory_base_model_sync() -> None:
    class Foo(BaseModel):
        bar: str
        baz: int

    def sync_fn() -> Foo:
        return Foo(bar="bar", baz=1)

    def handle_response_model(result, fn, context) -> None:
        assert isinstance(result, BaseModel)

    decorate = middleware_factory(handle_response_model=handle_response_model)(sync_fn)
    result = decorate()
    assert result.model_dump() == sync_fn().model_dump()


@pytest.mark.asyncio
async def test_middleware_factory_base_model_async() -> None:
    class Foo(BaseModel):
        bar: str
        baz: int

    async def async_fn() -> Foo:
        return Foo(bar="bar", baz=1)

    async def handle_response_model_async(result, fn, context) -> None:
        assert isinstance(result, BaseModel)

    decorate = middleware_factory(
        handle_response_model_async=handle_response_model_async
    )(async_fn)
    result = await decorate()
    async_fn_result = await async_fn()
    assert result.model_dump() == async_fn_result.model_dump()


def test_middleware_factory_structured_stream() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class Foo(BaseModel):
        bar: str
        baz: int

    mock_chunk = MagicMock()
    my_foo = Foo(bar="bar", baz=1)
    mock_chunk.content = my_foo.model_dump_json()
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    base_stream = MagicMock()
    base_stream.__iter__.return_value = (t for t in [(mock_chunk, None)])
    my_structured_stream = BaseStructuredStream(
        stream=base_stream, response_model=Foo, fields_from_call_args={}
    )

    def sync_fn() -> BaseStructuredStream:
        return my_structured_stream

    def handle_structured_stream(result, fn, context) -> None:
        assert isinstance(result, BaseStructuredStream)

    decorate = middleware_factory(handle_structured_stream=handle_structured_stream)(
        sync_fn
    )
    for chunk in decorate():
        assert chunk.model_dump() == my_foo.model_dump()


@pytest.mark.asyncio
async def test_middleware_factory_structured_stream_async() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class Foo(BaseModel):
        bar: str
        baz: int

    mock_chunk = MagicMock()
    my_foo = Foo(bar="bar", baz=1)
    mock_chunk.content = my_foo.model_dump_json()
    mock_chunk.input_tokens = 1
    mock_chunk.output_tokens = 2
    mock_chunk.model = "updated_model"

    base_stream = MagicMock()

    async def generator(self):
        yield mock_chunk, None

    base_stream.__aiter__ = generator
    my_structured_stream = BaseStructuredStream(
        stream=base_stream, response_model=Foo, fields_from_call_args={}
    )

    async def async_fn() -> BaseStructuredStream:
        return my_structured_stream

    async def handle_structured_stream_async(result, fn, context) -> None:
        assert isinstance(result, BaseStructuredStream)

    decorate = middleware_factory(
        handle_structured_stream_async=handle_structured_stream_async
    )(async_fn)
    async for chunk in await decorate():
        assert chunk.model_dump() == my_foo.model_dump()


def test_middleware_factory_handle_error_sync() -> None:
    def sync_fn():
        raise ValueError("An error occurred")

    def handle_error(e, fn, context):
        assert isinstance(e, ValueError)
        return "Handled Error"

    decorate = middleware_factory(handle_error=handle_error)(sync_fn)
    result = decorate()
    assert result == "Handled Error"


@pytest.mark.asyncio
async def test_middleware_factory_handle_error_async() -> None:
    async def async_fn():
        raise ValueError("An error occurred")

    async def handle_error_async(e, fn, context):
        assert isinstance(e, ValueError)
        return "Handled Error"

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    result = await decorate()
    assert result == "Handled Error"


def test_middleware_factory_handle_error_sync_exception() -> None:
    def sync_fn():
        raise ValueError("An error occurred")

    def handle_error(e, fn, context):
        raise RuntimeError("New error in handle_error")

    decorate = middleware_factory(handle_error=handle_error)(sync_fn)
    with pytest.raises(RuntimeError, match="New error in handle_error"):
        decorate()


@pytest.mark.asyncio
async def test_middleware_factory_handle_error_async_exception() -> None:
    async def async_fn():
        raise ValueError("An error occurred")

    async def handle_error_async(e, fn, context):
        raise RuntimeError("New error in handle_error_async")

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    with pytest.raises(RuntimeError, match="New error in handle_error_async"):
        await decorate()


def test_middleware_factory_handle_error_no_handler() -> None:
    def sync_fn():
        raise ValueError("An error occurred")

    decorate = middleware_factory()(sync_fn)
    with pytest.raises(ValueError, match="An error occurred"):
        decorate()


@pytest.mark.asyncio
async def test_middleware_factory_handle_error_async_no_handler() -> None:
    async def async_fn():
        raise ValueError("An error occurred")

    decorate = middleware_factory()(async_fn)
    with pytest.raises(ValueError, match="An error occurred"):
        await decorate()


def test_middleware_factory_stream_sync_exception() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    def sync_fn() -> BaseStream:
        def iterator():
            yield "chunk1", None
            raise ValueError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...  # pyright ignore [reportReturnType]

            def construct_call_response(
                self,
            ) -> BaseCallResponse: ...  # pyright ignore [reportReturnType]

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __iter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    handle_error_called = False

    def handle_error(e, fn, context):
        nonlocal handle_error_called
        handle_error_called = True
        assert isinstance(e, ValueError)
        return "Handled Stream Error"

    decorate = middleware_factory(handle_error=handle_error)(sync_fn)
    result = decorate()
    collected_chunks = []
    for chunk, _ in result:
        collected_chunks.append(chunk)
    assert collected_chunks == ["chunk1"]
    assert handle_error_called


@pytest.mark.asyncio
async def test_middleware_factory_stream_async_exception() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    async def async_fn() -> BaseStream:
        async def iterator():
            yield "chunk1", None
            raise ValueError("Stream error")

        class MyStream(BaseStream):
            def construct_call_response(self) -> BaseCallResponse: ...

            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __aiter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    handle_error_async_called = False

    async def handle_error_async(e, fn, context):
        nonlocal handle_error_async_called
        handle_error_async_called = True
        assert isinstance(e, ValueError)
        return "Handled Stream Error"

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    result = await decorate()
    collected_chunks = []
    async for chunk, _ in result:
        collected_chunks.append(chunk)
    assert collected_chunks == ["chunk1"]
    assert handle_error_async_called


def test_middleware_factory_stream_sync_exception_no_handler() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    def sync_fn() -> BaseStream:
        def iterator():
            yield "chunk1", None
            raise ValueError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            def construct_call_response(self) -> BaseCallResponse: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __iter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    decorate = middleware_factory()(sync_fn)
    result = decorate()
    with pytest.raises(ValueError, match="Stream error"):
        for chunk, _ in result:  # noqa: B007
            pass


@pytest.mark.asyncio
async def test_middleware_factory_stream_async_exception_no_handler() -> None:
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    async def async_fn() -> BaseStream:
        async def iterator():
            yield "chunk1", None
            raise ValueError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            def construct_call_response(self) -> BaseCallResponse: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __aiter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    decorate = middleware_factory()(async_fn)
    result = await decorate()
    with pytest.raises(ValueError, match="Stream error"):
        async for chunk, _ in result:  # noqa: B007
            pass


def test_middleware_factory_handle_error_sync_in_handle_call_response() -> None:
    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"  # pragma: no cover

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore

    def sync_fn() -> MyCallResponse:
        return call_response

    def handle_call_response(result, fn, context) -> None:
        raise RuntimeError("Error in handle_call_response")

    decorate = middleware_factory(handle_call_response=handle_call_response)(sync_fn)
    with pytest.raises(RuntimeError, match="Error in handle_call_response"):
        decorate()


@pytest.mark.asyncio
async def test_middleware_factory_handle_error_async_in_handle_call_response_async() -> (
    None
):
    class MyCallResponse(BaseCallResponse):
        @property
        def content(self) -> str:
            return "content"  # pragma: no cover

    patch.multiple(MyCallResponse, __abstractmethods__=set()).start()
    call_response = MyCallResponse(
        metadata={},
        response="",
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )  # type: ignore

    async def async_fn() -> MyCallResponse:
        return call_response

    async def handle_call_response_async(result, fn, context) -> None:
        raise RuntimeError("Error in handle_call_response_async")

    decorate = middleware_factory(
        handle_call_response_async=handle_call_response_async
    )(async_fn)
    with pytest.raises(RuntimeError, match="Error in handle_call_response_async"):
        await decorate()


def test_middleware_factory_handle_error_sync_in_custom_context_manager() -> None:
    def sync_fn():
        return "result"  # pragma: no cover

    def custom_context_manager(fn):
        class CustomContextManager:
            def __enter__(self):
                raise RuntimeError("Error in custom_context_manager")

            def __exit__(self, exc_type, exc_val, exc_tb):
                pass  # pragma: no cover

        return CustomContextManager()

    decorate = middleware_factory(custom_context_manager=custom_context_manager)(
        sync_fn
    )
    with pytest.raises(RuntimeError, match="Error in custom_context_manager"):
        decorate()


@pytest.mark.asyncio
async def test_middleware_factory_handle_error_async_in_custom_context_manager() -> (
    None
):
    async def async_fn():
        return "result"  # pragma: no cover

    def custom_context_manager(fn):
        class CustomContextManager:
            def __enter__(self):
                raise RuntimeError("Error in custom_context_manager")

            def __exit__(self, exc_type, exc_val, exc_tb):
                pass  # pragma: no cover

        return CustomContextManager()

    decorate = middleware_factory(custom_context_manager=custom_context_manager)(
        async_fn
    )
    with pytest.raises(RuntimeError, match="Error in custom_context_manager"):
        await decorate()


def test_middleware_factory_no_handle_functions() -> None:
    def sync_fn():
        return "result"

    decorate = middleware_factory()(sync_fn)
    result = decorate()
    assert result == "result"


@pytest.mark.asyncio
async def test_middleware_factory_no_handle_functions_async() -> None:
    async def async_fn():
        return "result"

    decorate = middleware_factory()(async_fn)
    result = await decorate()
    assert result == "result"


def test_middleware_factory_handle_error_in_handle_error() -> None:
    def sync_fn():
        raise ValueError("An error occurred")

    def handle_error(e, fn, context):
        raise RuntimeError("Error in handle_error")

    decorate = middleware_factory(handle_error=handle_error)(sync_fn)
    with pytest.raises(RuntimeError, match="Error in handle_error"):
        decorate()


@pytest.mark.asyncio
async def test_middleware_factory_handle_error_async_in_handle_error_async() -> None:
    async def async_fn():
        raise ValueError("An error occurred")

    async def handle_error_async(e, fn, context):
        raise RuntimeError("Error in handle_error_async")

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    with pytest.raises(RuntimeError, match="Error in handle_error_async"):
        await decorate()


@pytest.mark.asyncio
async def test_middleware_factory_stream_async_with_error() -> None:
    """Test error handling in async stream."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    async def async_fn() -> BaseStream:
        async def iterator():
            yield "chunk1", None
            raise StreamError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            def construct_call_response(self) -> BaseCallResponse: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __aiter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    error_handled = False

    async def handle_error_async(e, fn, context):
        nonlocal error_handled
        error_handled = True
        assert isinstance(e, StreamError)
        return "Error Handled"

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    result = await decorate()
    chunks = []
    async for chunk, _ in result:
        chunks.append(chunk)

    assert chunks == ["chunk1"]
    assert error_handled


@pytest.mark.asyncio
async def test_middleware_factory_stream_async_error_handling_fails() -> None:
    """Test when error handler itself raises an exception."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    async def async_fn() -> BaseStream:
        async def iterator():
            yield "chunk1", None
            raise StreamError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            def construct_call_response(self) -> BaseCallResponse: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __aiter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    async def handle_error_async(e, fn, context):
        raise RuntimeError("Error handler failed")

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    result = await decorate()

    with pytest.raises(RuntimeError, match="Error handler failed"):
        async for _ in result:
            pass


def test_middleware_factory_structured_stream_with_error() -> None:
    """Test error handling in structured stream."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class Foo(BaseModel):
        bar: str
        baz: int

    def sync_fn() -> BaseStructuredStream:
        def iterator():
            yield "chunk1"
            raise StreamError("Stream error")

        class MyStructuredStream(BaseStructuredStream):
            def __iter__(self):
                return iterator()

        base_stream = MagicMock()
        base_stream.__iter__.return_value = iterator()
        return MyStructuredStream(
            stream=base_stream, response_model=Foo, fields_from_call_args={}
        )

    error_handled = False

    def handle_error(e, fn, context):
        nonlocal error_handled
        error_handled = True
        assert isinstance(e, StreamError)
        return "Error Handled"

    decorate = middleware_factory(handle_error=handle_error)(sync_fn)
    result = decorate()
    chunks = []
    for chunk in result:
        chunks.append(chunk)

    assert chunks == ["chunk1"]
    assert error_handled


@pytest.mark.asyncio
async def test_middleware_factory_structured_stream_async_with_error() -> None:
    """Test error handling in async structured stream."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class Foo(BaseModel):
        bar: str
        baz: int

    async def async_fn() -> BaseStructuredStream:
        async def iterator():
            yield "chunk1"
            raise StreamError("Stream error")

        class MyStructuredStream(BaseStructuredStream):
            def __aiter__(self):
                return iterator()

        base_stream = MagicMock()
        base_stream.__aiter__ = iterator
        return MyStructuredStream(
            stream=base_stream, response_model=Foo, fields_from_call_args={}
        )

    error_handled = False

    async def handle_error_async(e, fn, context):
        nonlocal error_handled
        error_handled = True
        assert isinstance(e, StreamError)
        return "Error Handled"

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    result = await decorate()
    chunks = []
    async for chunk in result:
        chunks.append(chunk)

    assert chunks == ["chunk1"]
    assert error_handled


def test_stream_without_error_handler() -> None:
    """Test stream behavior when no error handler is provided."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    def sync_fn() -> BaseStream:
        def iterator():
            yield "chunk1", None
            raise StreamError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            def construct_call_response(self) -> BaseCallResponse: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __iter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    decorate = middleware_factory()(sync_fn)
    result = decorate()
    chunks = []

    with pytest.raises(StreamError, match="Stream error"):
        for chunk, _ in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]


def test_structured_stream_without_error_handler() -> None:
    """Test structured stream behavior when no error handler is provided."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class Foo(BaseModel):
        bar: str
        baz: int

    def sync_fn() -> BaseStructuredStream:
        def iterator():
            yield "chunk1"
            raise StreamError("Stream error")

        class MyStructuredStream(BaseStructuredStream):
            def __iter__(self):
                return iterator()

        base_stream = MagicMock()
        base_stream.__iter__.return_value = iterator()
        return MyStructuredStream(
            stream=base_stream, response_model=Foo, fields_from_call_args={}
        )

    decorate = middleware_factory()(sync_fn)
    result = decorate()
    chunks = []

    with pytest.raises(StreamError, match="Stream error"):
        for chunk in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]


@pytest.mark.asyncio
async def test_middleware_factory_stream_error_handler_exception_async() -> None:
    """Test when stream error handler raises an exception in async stream."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class ErrorHandlerError(Exception):
        pass

    handled = False

    async def async_fn() -> BaseStream:
        async def iterator():
            yield "chunk1", None
            raise StreamError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            def construct_call_response(self) -> BaseCallResponse: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __aiter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    async def handle_error_async(e, fn, context):
        nonlocal handled
        handled = True
        assert isinstance(e, StreamError)
        raise ErrorHandlerError("Error handler failed")

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    result = await decorate()
    chunks = []

    with pytest.raises(ErrorHandlerError, match="Error handler failed"):
        async for chunk, _ in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]
    assert handled


@pytest.mark.asyncio
async def test_middleware_factory_structured_stream_error_handler_exception_async() -> (
    None
):
    """Test when structured stream error handler raises an exception in async stream."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class ErrorHandlerError(Exception):
        pass

    class Foo(BaseModel):
        bar: str
        baz: int

    handled = False

    async def async_fn() -> BaseStructuredStream:
        async def iterator():
            yield "chunk1"
            raise StreamError("Stream error")

        class MyStructuredStream(BaseStructuredStream):
            def __aiter__(self):
                return iterator()

        base_stream = MagicMock()
        base_stream.__aiter__ = iterator
        return MyStructuredStream(
            stream=base_stream, response_model=Foo, fields_from_call_args={}
        )

    async def handle_error_async(e, fn, context):
        nonlocal handled
        handled = True
        assert isinstance(e, StreamError)
        raise ErrorHandlerError("Error handler failed")

    decorate = middleware_factory(handle_error_async=handle_error_async)(async_fn)
    result = await decorate()
    chunks = []

    with pytest.raises(ErrorHandlerError, match="Error handler failed"):
        async for chunk in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]
    assert handled


def test_middleware_factory_stream_error_handler_exception() -> None:
    """Test when stream error handler raises an exception in sync stream."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class ErrorHandlerError(Exception):
        pass

    handled = False

    def sync_fn() -> BaseStream:
        def iterator():
            yield "chunk1", None
            raise StreamError("Stream error")

        class MyStream(BaseStream):
            def _construct_message_param(
                self, tool_calls: list[Any] | None = None, content: str | None = None
            ) -> AssistantMessageTypeDef: ...

            def construct_call_response(self) -> BaseCallResponse: ...

            @property
            def cost(self) -> int:
                return 10  # pragma: no cover

            def __iter__(self):
                return iterator()

            @property
            def cost_metadata(self) -> CostMetadata: ...

        return MyStream(
            stream=None,  # pyright: ignore [reportArgumentType]
            metadata={},
            tool_types=[],
            call_response_type=MagicMock,
            model="model",
            prompt_template="prompt_template",
            fn_args={},
            dynamic_config=None,
            messages=[],
            call_params={},
            call_kwargs={},
        )

    def handle_error(e, fn, context):
        nonlocal handled
        handled = True
        assert isinstance(e, StreamError)
        raise ErrorHandlerError("Error handler failed")

    decorate = middleware_factory(handle_error=handle_error)(sync_fn)
    result = decorate()
    chunks = []

    with pytest.raises(ErrorHandlerError, match="Error handler failed"):
        for chunk, _ in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]
    assert handled


def test_middleware_factory_structured_stream_error_handler_exception() -> None:
    """Test when structured stream error handler raises an exception in sync stream."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class ErrorHandlerError(Exception):
        pass

    class Foo(BaseModel):
        bar: str
        baz: int

    handled = False

    def sync_fn() -> BaseStructuredStream:
        def iterator():
            yield "chunk1"
            raise StreamError("Stream error")

        class MyStructuredStream(BaseStructuredStream):
            def __iter__(self):
                return iterator()

        base_stream = MagicMock()
        base_stream.__iter__.return_value = iterator()
        return MyStructuredStream(
            stream=base_stream, response_model=Foo, fields_from_call_args={}
        )

    def handle_error(e, fn, context):
        nonlocal handled
        handled = True
        assert isinstance(e, StreamError)
        raise ErrorHandlerError("Error handler failed")

    decorate = middleware_factory(handle_error=handle_error)(sync_fn)
    result = decorate()
    chunks = []

    with pytest.raises(ErrorHandlerError, match="Error handler failed"):
        for chunk in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]
    assert handled


def test_middleware_factory_structured_stream_error_no_handler() -> None:
    """Test when structured stream raises an error and no handler is provided."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class Foo(BaseModel):
        bar: str
        baz: int

    def sync_fn() -> BaseStructuredStream:
        def iterator():
            yield "chunk1"
            raise StreamError("Stream error")

        class MyStructuredStream(BaseStructuredStream):
            def __iter__(self):
                return iterator()

        base_stream = MagicMock()
        base_stream.__iter__.return_value = iterator()
        return MyStructuredStream(
            stream=base_stream, response_model=Foo, fields_from_call_args={}
        )

    decorate = middleware_factory()(sync_fn)
    result = decorate()
    chunks = []

    with pytest.raises(StreamError, match="Stream error"):
        for chunk in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]


@pytest.mark.asyncio
async def test_middleware_factory_structured_stream_error_no_handler_async() -> None:
    """Test when async structured stream raises an error and no handler is provided."""
    patch.multiple(BaseStream, __abstractmethods__=set()).start()

    class StreamError(Exception):
        pass

    class Foo(BaseModel):
        bar: str
        baz: int

    async def async_fn() -> BaseStructuredStream:
        async def iterator():
            yield "chunk1"
            raise StreamError("Stream error")

        class MyStructuredStream(BaseStructuredStream):
            def __aiter__(self):
                return iterator()

        base_stream = MagicMock()
        base_stream.__aiter__ = iterator
        return MyStructuredStream(
            stream=base_stream, response_model=Foo, fields_from_call_args={}
        )

    decorate = middleware_factory()(async_fn)
    result = await decorate()
    chunks = []

    with pytest.raises(StreamError, match="Stream error"):
        async for chunk in result:
            chunks.append(chunk)

    assert chunks == ["chunk1"]



"""The `google_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import GoogleCallParams
from .call_response import GoogleCallResponse
from .call_response_chunk import GoogleCallResponseChunk
from .stream import GoogleStream
from .tool import GoogleTool

google_call = call_factory(
    TCallResponse=GoogleCallResponse,
    TCallResponseChunk=GoogleCallResponseChunk,
    TStream=GoogleStream,
    TToolType=GoogleTool,
    default_call_params=GoogleCallParams(),
    setup_call=setup_call,
    get_json_output=get_json_output,
    handle_stream=handle_stream,  # pyright: ignore [reportArgumentType]
    handle_stream_async=handle_stream_async,  # pyright: ignore [reportArgumentType]
)
"""A decorator for calling the Google API with a typed function.

usage docs: learn/calls.md

This decorator is used to wrap a typed function that calls the Google API. It parses
the prompt template of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.google import google_call


@google_call("google-1.5-flash")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"

response = recommend_book("fantasy")
print(response.content)
```

Args:
    model (str): The Google model to use in the API call.
    stream (bool): Whether to stream the response from the API call.
    tools (list[BaseTool | Callable]): The tools to use in the Google API call.
    response_model (BaseModel | BaseType): The response model into which the response
        should be structured.
    output_parser (Callable[[GoogleCallResponse | ResponseModelT], Any]): A function
        for parsing the call response whose value will be returned in place of the
        original call response.
    json_modem (bool): Whether to use JSON Mode.
    client (object): An optional custom client to use in place of the default client.
    call_params (GoogleCallParams): The `GoogleCallParams` call parameters to use in the
        API call.

Returns:
    decorator (Callable): The decorator for turning a typed function into a Google API
        call.
"""



"""Tests the `call_factory` module."""

from unittest.mock import MagicMock, patch

import pytest

from mirascope.core.base._call_factory import call_factory


@pytest.fixture()
def mock_call_factory_kwargs() -> dict:
    """Returns the mock kwargs for the `call_factory` function."""
    return {
        "TCallResponse": MagicMock,
        "TCallResponseChunk": MagicMock,
        "TToolType": MagicMock,
        "TStream": MagicMock,
        "default_call_params": MagicMock(),
        "setup_call": MagicMock(),
        "get_json_output": MagicMock(),
        "handle_stream": MagicMock(),
        "handle_stream_async": MagicMock(),
    }


@patch("mirascope.core.base._call_factory.create_factory", new_callable=MagicMock)
@patch("mirascope.core.base._call_factory.partial", new_callable=MagicMock)
def test_call_factory_create(
    mock_partial: MagicMock,
    mock_create_factory: MagicMock,
    mock_call_factory_kwargs: dict,
) -> None:
    """Tests the `create_factory` route for the `call_factory` method."""
    call = call_factory(**mock_call_factory_kwargs)
    create_kwargs = {
        "model": "model",
        "tools": [],
        "response_model": None,
        "output_parser": MagicMock(),
        "json_mode": True,
        "client": MagicMock(),
    }
    _ = call(**create_kwargs)
    mock_create_factory.assert_called_once_with(
        TCallResponse=mock_call_factory_kwargs["TCallResponse"],
        setup_call=mock_call_factory_kwargs["setup_call"],
    )
    mock_partial.assert_called_once_with(
        mock_create_factory.return_value,
        **create_kwargs,
        call_params=mock_call_factory_kwargs["default_call_params"],
    )


@patch("mirascope.core.base._call_factory.stream_factory", new_callable=MagicMock)
@patch("mirascope.core.base._call_factory.partial", new_callable=MagicMock)
def test_call_factory_stream(
    mock_partial: MagicMock,
    mock_stream_factory: MagicMock,
    mock_call_factory_kwargs: dict,
) -> None:
    """Tests the `stream_factory` route for the `call_factory` method."""
    call = call_factory(**mock_call_factory_kwargs)
    stream_kwargs = {
        "model": "model",
        "tools": [],
        "json_mode": False,
        "client": MagicMock(),
        "call_params": MagicMock(),
    }
    _ = call(stream=True, **stream_kwargs)
    mock_stream_factory.assert_called_once_with(
        TCallResponse=mock_call_factory_kwargs["TCallResponse"],
        TStream=mock_call_factory_kwargs["TStream"],
        setup_call=mock_call_factory_kwargs["setup_call"],
        handle_stream=mock_call_factory_kwargs["handle_stream"],
        handle_stream_async=mock_call_factory_kwargs["handle_stream_async"],
    )
    mock_partial.assert_called_once_with(
        mock_stream_factory.return_value, **stream_kwargs, partial_tools=False
    )


@patch("mirascope.core.base._call_factory.extract_factory", new_callable=MagicMock)
@patch("mirascope.core.base._call_factory.partial", new_callable=MagicMock)
def test_call_factory_extract(
    mock_partial: MagicMock,
    mock_extract_factory: MagicMock,
    mock_call_factory_kwargs: dict,
) -> None:
    """Tests the `extract_factory` route for the `call_factory` method."""
    call = call_factory(**mock_call_factory_kwargs)
    extract_kwargs = {
        "model": "model",
        "response_model": MagicMock,
        "output_parser": None,
        "json_mode": False,
        "client": MagicMock(),
        "call_params": MagicMock(),
    }
    _ = call(**extract_kwargs)
    mock_extract_factory.assert_called_once_with(
        TCallResponse=mock_call_factory_kwargs["TCallResponse"],
        TToolType=mock_call_factory_kwargs["TToolType"],
        setup_call=mock_call_factory_kwargs["setup_call"],
        get_json_output=mock_call_factory_kwargs["get_json_output"],
    )
    mock_partial.assert_called_once_with(
        mock_extract_factory.return_value, **extract_kwargs
    )


@patch(
    "mirascope.core.base._call_factory.extract_with_tools_factory",
    new_callable=MagicMock,
)
@patch("mirascope.core.base._call_factory.partial", new_callable=MagicMock)
def test_call_factory_extract_with_tools(
    mock_partial: MagicMock,
    mock_extract_with_tools_factory: MagicMock,
    mock_call_factory_kwargs: dict,
) -> None:
    """Tests the `extract_factory` route for the `call_factory` method."""
    call = call_factory(**mock_call_factory_kwargs)
    extract_with_tools_kwargs = {
        "model": "model",
        "tools": [MagicMock()],
        "response_model": MagicMock,
        "output_parser": None,
        "client": MagicMock(),
        "call_params": MagicMock(),
    }
    _ = call(**extract_with_tools_kwargs)
    mock_extract_with_tools_factory.assert_called_once_with(
        TCallResponse=mock_call_factory_kwargs["TCallResponse"],
        setup_call=mock_call_factory_kwargs["setup_call"],
        get_json_output=mock_call_factory_kwargs["get_json_output"],
    )
    mock_partial.assert_called_once_with(
        mock_extract_with_tools_factory.return_value,
        **extract_with_tools_kwargs,
    )


@patch(
    "mirascope.core.base._call_factory.structured_stream_factory",
    new_callable=MagicMock,
)
@patch("mirascope.core.base._call_factory.partial", new_callable=MagicMock)
def test_call_factory_structured_stream(
    mock_partial: MagicMock,
    mock_structured_stream_factory: MagicMock,
    mock_call_factory_kwargs: dict,
) -> None:
    """Tests the `structured_stream_factory` route for the `call_factory` method."""
    call = call_factory(**mock_call_factory_kwargs)
    structured_stream_kwargs = {
        "model": "model",
        "response_model": MagicMock,
        "json_mode": False,
        "client": MagicMock(),
        "call_params": MagicMock(),
    }
    _ = call(stream=True, **structured_stream_kwargs)
    mock_structured_stream_factory.assert_called_once_with(
        TCallResponse=mock_call_factory_kwargs["TCallResponse"],
        TCallResponseChunk=mock_call_factory_kwargs["TCallResponseChunk"],
        TStream=mock_call_factory_kwargs["TStream"],
        TToolType=mock_call_factory_kwargs["TToolType"],
        setup_call=mock_call_factory_kwargs["setup_call"],
        get_json_output=mock_call_factory_kwargs["get_json_output"],
    )
    mock_partial.assert_called_once_with(
        mock_structured_stream_factory.return_value, **structured_stream_kwargs
    )


def test_call_decorator_invalid_output_parser_with_stream(
    mock_call_factory_kwargs: dict,
) -> None:
    """Tests a ValueError is raised if `output_parser` is provided and `stream=True`."""
    call = call_factory(**mock_call_factory_kwargs)
    with pytest.raises(
        ValueError, match="Cannot use `output_parser` with `stream=True`"
    ):
        call("model", stream=True, output_parser=MagicMock())



"""Tests the `google.stream` module."""

import pytest
from google.genai.types import (
    Candidate,
    Content,
    ContentDict,
    GenerateContentResponse,
    GenerateContentResponseUsageMetadata,
    Part,
    PartDict,
)
from google.genai.types import (
    FinishReason as GoogleFinishReason,
)

from mirascope.core.base.types import CostMetadata
from mirascope.core.google.call_response import GoogleCallResponse
from mirascope.core.google.call_response_chunk import GoogleCallResponseChunk
from mirascope.core.google.stream import GoogleStream


def test_google_stream() -> None:
    """Tests the `GoogleStream` class."""
    assert GoogleStream._provider == "google"

    chunks = [
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(parts=[Part(text="The author is ")], role="model"),
                )
            ]
        ),
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(
                        parts=[Part(text="Patrick Rothfuss")], role="model"
                    ),
                )
            ]
        ),
    ]
    stream = GoogleStream(
        stream=((GoogleCallResponseChunk(chunk=chunk), None) for chunk in chunks),
        metadata={},
        tool_types=None,
        call_response_type=GoogleCallResponse,
        model="google-1.5-flash",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "parts": [{"text": "Who is the author?"}]}],
        call_params={},
        call_kwargs={},
    )

    with pytest.raises(
        ValueError, match="No stream response, check if the stream has been consumed."
    ):
        stream.construct_call_response()

    assert stream.cost is None
    for _ in stream:
        pass
    assert stream.cost is None
    assert stream.cost_metadata == CostMetadata()
    assert stream.message_param == {
        "role": "model",
        "parts": [{"text": "The author is Patrick Rothfuss"}],
    }


def test_construct_call_response() -> None:
    chunks = [
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(parts=[Part(text="The author is ")], role="model"),
                )
            ]
        ),
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=GoogleFinishReason.STOP,
                    content=Content(
                        parts=[Part(text="Patrick Rothfuss")], role="model"
                    ),
                )
            ]
        ),
    ]
    stream = GoogleStream(
        stream=((GoogleCallResponseChunk(chunk=chunk), None) for chunk in chunks),
        metadata={},
        tool_types=None,
        call_response_type=GoogleCallResponse,
        model="google-1.5-flash",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[
            ContentDict(role="user", parts=[PartDict(text="Who is the author?")])
        ],
        call_params={},
        call_kwargs={},
    )

    for _ in stream:
        pass

    response = GenerateContentResponse(
        candidates=[
            Candidate(
                finish_reason=GoogleFinishReason.STOP,
                content=Content(
                    parts=[Part(text="The author is Patrick Rothfuss")],
                    role="model",
                ),
            )
        ],
        model_version="google-1.5-flash",
        usage_metadata=GenerateContentResponseUsageMetadata(
            candidates_token_count=None, prompt_token_count=None, total_token_count=0
        ),
    )
    call_response = GoogleCallResponse(
        metadata={},
        response=response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param={},
        start_time=0,
        end_time=0,
    )
    constructed_call_response = stream.construct_call_response()
    assert constructed_call_response.response.__eq__(call_response.response)



from typing import cast

from ...base.call_params import CommonCallParams
from ..call_params import AnthropicCallParams

ANTHROPIC_PARAM_MAPPING = {
    "temperature": "temperature",
    "max_tokens": "max_tokens",
    "top_p": "top_p",
    "stop": "stop_sequences",
}


def convert_common_call_params(common_params: CommonCallParams) -> AnthropicCallParams:
    """Convert CommonCallParams to Anthropic parameters."""
    return cast(
        AnthropicCallParams,
        {"max_tokens": 1024}
        | {
            ANTHROPIC_PARAM_MAPPING[key]: [value] if isinstance(value, str) else value
            for key, value in common_params.items()
            if key in ANTHROPIC_PARAM_MAPPING
            and ANTHROPIC_PARAM_MAPPING[key] is not None
        },
    )



"""Tests the `xai._utils.setup_call` module."""

from unittest.mock import MagicMock, patch

from mirascope.core.xai._utils._setup_call import setup_call


@patch("mirascope.core.xai._utils._setup_call.OpenAI", new_callable=MagicMock)
@patch(
    "mirascope.core.xai._utils._setup_call.setup_call_openai",
    new_callable=MagicMock,
)
def test_setup_call(
    mock_setup_call_openai: MagicMock,
    mock_openai: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_setup_call_openai.return_value = [MagicMock() for _ in range(5)]
    mock_openai.return_value = MagicMock()
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="gpt-4o",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    mock_setup_call_openai.assert_called_once_with(
        model="gpt-4o",
        client=mock_openai.return_value,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert prompt_template == mock_setup_call_openai.return_value[1]
    assert messages == mock_setup_call_openai.return_value[2]
    assert tool_types == mock_setup_call_openai.return_value[3]
    assert call_kwargs == mock_setup_call_openai.return_value[4]



from mirascope.core.base.tool import BaseTool
from mirascope.llm.tool import Tool


class DummyTool(BaseTool):
    def call(self):
        return "dummy_tool_call"

    @property
    def model_fields(self):  # pyright: ignore [reportIncompatibleMethodOverride,reportIncompatibleVariableOverride]
        return ["field1"]

    field1: str = "value"


def test_tool():
    dummy_tool_instance = DummyTool()
    tool_instance = Tool(tool=dummy_tool_instance)  # pyright: ignore [reportAbstractUsage]
    assert tool_instance.call() == "dummy_tool_call"
    assert tool_instance.field1 == "value"



from typing import TypeAlias

from .. import BaseMessageParam
from ._call import bedrock_call
from ._call import bedrock_call as call
from ._types import (
    AssistantMessageTypeDef,
    InternalBedrockMessageParam,
    UserMessageTypeDef,
)
from .call_params import BedrockCallParams
from .call_response import BedrockCallResponse
from .call_response_chunk import BedrockCallResponseChunk
from .dynamic_config import AsyncBedrockDynamicConfig, BedrockDynamicConfig
from .stream import BedrockStream
from .tool import BedrockTool, BedrockToolConfig

BedrockMessageParam: TypeAlias = InternalBedrockMessageParam | BaseMessageParam

__all__ = [
    "AssistantMessageTypeDef",
    "AsyncBedrockDynamicConfig",
    "BedrockCallParams",
    "BedrockCallResponse",
    "BedrockCallResponseChunk",
    "BedrockDynamicConfig",
    "BedrockMessageParam",
    "BedrockStream",
    "BedrockTool",
    "BedrockToolConfig",
    "UserMessageTypeDef",
    "bedrock_call",
    "call",
]



"""Handles streaming content and tools from the Google API."""

from collections.abc import AsyncGenerator, Generator

from google.genai.types import GenerateContentResponse

from ..call_response_chunk import GoogleCallResponseChunk
from ..tool import GoogleTool


def _handle_chunk(
    chunk: GenerateContentResponse, tool_types: list[type[GoogleTool]] | None = None
) -> Generator[tuple[GoogleCallResponseChunk, GoogleTool | None], None, None]:
    """Handles a chunk of the stream and yields any tools that are found.

    Note: google seems to only ever return full tool calls, so no partial possible here.
    """
    call_response_chunk = GoogleCallResponseChunk(chunk=chunk)
    has_tools = False
    if tool_types and (
        (candidates := chunk.candidates)
        and (content := candidates[0].content)
        and (parts := content.parts)
    ):
        for part in parts:
            if function_call := part.function_call:
                has_tools = True
                for tool_type in tool_types:
                    if tool_type._name() == function_call.name:
                        tool = tool_type.from_tool_call(function_call)
                        yield (call_response_chunk, tool)
    if not has_tools:
        yield call_response_chunk, None


def handle_stream(
    stream: Generator[GenerateContentResponse, None, None],
    tool_types: list[type[GoogleTool]] | None = None,
    partial_tools: bool = False,
) -> Generator[tuple[GoogleCallResponseChunk, GoogleTool | None], None, None]:
    """Iterator over the stream and constructs tools as they are streamed."""
    for chunk in stream:
        yield from _handle_chunk(chunk, tool_types)


async def handle_stream_async(
    stream: AsyncGenerator[GenerateContentResponse, None],
    tool_types: list[type[GoogleTool]] | None = None,
    partial_tools: bool = False,
) -> AsyncGenerator[tuple[GoogleCallResponseChunk, GoogleTool | None], None]:
    """
    Async iterator over the stream and constructs tools as they are streamed.

    Note: google does not currently support streaming tools.
    """
    async for chunk in stream:
        for call_response_chunk, tool in _handle_chunk(chunk, tool_types):
            yield call_response_chunk, tool



"""This module defines the function return type for functions as LLM calls."""

from typing import TypeAlias

from ..openai import AsyncOpenAIDynamicConfig, OpenAIDynamicConfig

AsyncXAIDynamicConfig: TypeAlias = AsyncOpenAIDynamicConfig
XAIDynamicConfig: TypeAlias = OpenAIDynamicConfig



"""This module contains the base call response class."""

from __future__ import annotations

import base64
import json
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from functools import cached_property, wraps
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeVar, cast

from pydantic import (
    BaseModel,
    ConfigDict,
    FieldSerializationInfo,
    SkipValidation,
    computed_field,
    field_serializer,
)

from ..costs import calculate_cost
from ._utils import BaseMessageParamConverter, BaseType, get_common_usage
from .call_kwargs import BaseCallKwargs
from .call_params import BaseCallParams
from .dynamic_config import BaseDynamicConfig
from .metadata import Metadata
from .tool import BaseTool
from .types import CostMetadata, FinishReason, JsonableType, Provider, Usage

if TYPE_CHECKING:
    from ...llm.tool import Tool
    from .. import BaseMessageParam

_ResponseT = TypeVar("_ResponseT", bound=Any)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ToolSchemaT = TypeVar("_ToolSchemaT")
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_MessageParamT = TypeVar("_MessageParamT", bound=Any)
_ToolMessageParamT = TypeVar("_ToolMessageParamT", bound=Any)
_CallParamsT = TypeVar("_CallParamsT", bound=BaseCallParams)
_UserMessageParamT = TypeVar("_UserMessageParamT", bound=Any)
_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound="BaseCallResponse")
_BaseMessageParamConverterT = TypeVar(
    "_BaseMessageParamConverterT", bound=BaseMessageParamConverter
)


def transform_tool_outputs(
    fn: Callable[
        [type[_BaseCallResponseT], Sequence[tuple[_BaseToolT, str]]],
        list[_ToolMessageParamT],
    ],
) -> Callable[
    [type[_BaseCallResponseT], Sequence[tuple[_BaseToolT, JsonableType]]],
    list[_ToolMessageParamT],
]:
    @wraps(fn)
    def wrapper(
        cls: type[_BaseCallResponseT],
        tools_and_outputs: Sequence[tuple[_BaseToolT, JsonableType]],
    ) -> list[_ToolMessageParamT]:
        def recursive_serializer(value: JsonableType) -> BaseType:
            if isinstance(value, str | int | float | bool | None):
                return value
            if isinstance(value, bytes):
                return base64.b64encode(value).decode("utf-8")
            if isinstance(value, BaseModel):
                return value.model_dump()
            if isinstance(value, list | set | tuple):
                return [recursive_serializer(item) for item in value]
            if isinstance(value, dict):
                return {k: recursive_serializer(v) for k, v in value.items()}
            raise TypeError(f"Unsupported type for serialization: {type(value)}")

        transformed_tools_and_outputs = [
            (
                tool,
                output.model_dump_json()
                if isinstance(output, BaseModel)
                else str(recursive_serializer(output))
                if isinstance(output, str | bytes)
                else json.dumps(recursive_serializer(output)),
            )
            for tool, output in tools_and_outputs
        ]
        return fn(cls, transformed_tools_and_outputs)

    return wrapper


class BaseCallResponse(
    BaseModel,
    Generic[
        _ResponseT,
        _BaseToolT,
        _ToolSchemaT,
        _BaseDynamicConfigT,
        _MessageParamT,
        _CallParamsT,
        _UserMessageParamT,
        _BaseMessageParamConverterT,
    ],
    ABC,
):
    """A base abstract interface for LLM call responses.

    Attributes:
        metadata: The metadata pulled from the call that was made.
        response: The original response from whichever model response this wraps.
        tool_types: The list of tool types used, if any.
        prompt_template: The unformatted prompt template from the call that was made.
        fn_args: The input arguments used when making the call.
        dynamic_config: Dynamic configuration options, if any.
        messages: The list of provider-specific messages used to make the API call.
        call_params: The original call params set in the call decorator.
        call_kwargs: The keyword arguments used to make the API call.
        user_message_param: The most recent provider-specific message if it was a user
            message. Otherwise `None`.
        start_time: The start time of the completion in ms.
        end_time: The end time of the completion in ms.
    """

    metadata: Metadata
    response: _ResponseT
    tool_types: list[type[_BaseToolT]] | None = None
    prompt_template: str | None
    fn_args: dict[str, Any]
    dynamic_config: _BaseDynamicConfigT
    messages: SkipValidation[list[_MessageParamT]]
    call_params: SkipValidation[_CallParamsT]
    call_kwargs: BaseCallKwargs[_ToolSchemaT]
    user_message_param: _UserMessageParamT | None = None
    start_time: float
    end_time: float

    _message_converter: type[_BaseMessageParamConverterT]
    _provider: ClassVar[str] = "NO PROVIDER"
    _model: str = "NO MODEL"

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    @field_serializer("tool_types", when_used="json")
    def serialize_tool_types(
        self, tool_types: list[type[_BaseToolT]] | None, info: FieldSerializationInfo
    ) -> list[dict[str, str]]:
        return [{"type": "function", "name": tool._name()} for tool in tool_types or []]

    def __str__(self) -> str:
        """Returns the string content of the response."""
        return self.content

    @computed_field
    @property
    @abstractmethod
    def content(self) -> str:
        """Should return the string content of the response.

        If there are multiple choices in a response, this method should select the 0th
        choice and return it's string content.

        If there is no string content (e.g. when using tools), this method must return
        the empty string.
        """
        ...

    @computed_field
    @property
    @abstractmethod
    def finish_reasons(self) -> list[str] | None:
        """Should return the finish reasons of the response.

        If there is no finish reason, this method must return None.
        """
        ...

    @computed_field
    @property
    @abstractmethod
    def model(self) -> str | None:
        """Should return the name of the response model."""
        ...

    @computed_field
    @property
    @abstractmethod
    def id(self) -> str | None:
        """Should return the id of the response."""
        ...

    @property
    @abstractmethod
    def usage(self) -> Any:  # noqa: ANN401
        """Should return the usage of the response.

        If there is no usage, this method must return None.
        """
        ...

    @computed_field
    @property
    @abstractmethod
    def input_tokens(self) -> int | float | None:
        """Should return the number of input tokens.

        If there is no input_tokens, this method must return None.
        """
        ...

    @computed_field
    @property
    @abstractmethod
    def cached_tokens(self) -> int | float | None:
        """Should return the number of cached tokens.

        If there is no cached_tokens, this method must return None.
        """
        ...

    @computed_field
    @property
    @abstractmethod
    def output_tokens(self) -> int | float | None:
        """Should return the number of output tokens.

        If there is no output_tokens, this method must return None.
        """
        ...

    @computed_field
    @property
    @abstractmethod
    def cost_metadata(self) -> CostMetadata:
        """Get metadata required for cost calculation.

        Returns:
            Metadata relevant to cost calculation
        """

        return CostMetadata(
            input_tokens=self.input_tokens,
            output_tokens=self.output_tokens,
            cached_tokens=self.cached_tokens,
        )

    @computed_field
    @property
    def cost(self) -> float | None:
        """Calculate the cost of this API call using the unified calculate_cost function."""

        model = self.model
        if not model:
            return None

        if self.input_tokens is None or self.output_tokens is None:
            return None

        return calculate_cost(
            provider=self.provider,
            model=model,
            metadata=self.cost_metadata,
        )

    @property
    def provider(self) -> Provider:
        """Get the provider used for this API call."""
        return cast(Provider, self._provider)

    @computed_field
    @cached_property
    @abstractmethod
    def message_param(self) -> Any:  # noqa: ANN401
        """Returns the assistant's response as a message parameter."""
        ...

    @cached_property
    @abstractmethod
    def tools(self) -> list[_BaseToolT] | None:
        """Returns the tools for the 0th choice message."""
        ...

    @cached_property
    @abstractmethod
    def tool(self) -> _BaseToolT | None:
        """Returns the 0th tool for the 0th choice message."""
        ...

    @classmethod
    @abstractmethod
    @transform_tool_outputs
    def tool_message_params(
        cls, tools_and_outputs: Sequence[tuple[_BaseToolT, str]]
    ) -> list[Any]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The sequence of tools and their outputs from which the tool
                message parameters should be constructed.
        """
        ...

    @property
    @abstractmethod
    def common_finish_reasons(self) -> list[FinishReason] | None:
        """Provider-agnostic finish reasons."""
        ...

    @property
    @abstractmethod
    def common_message_param(self) -> BaseMessageParam:
        """Provider-agnostic assistant message param."""
        ...

    @property
    @abstractmethod
    def common_user_message_param(self) -> BaseMessageParam | None:
        """Provider-agnostic user message param."""
        ...

    @property
    def common_messages(self) -> list[BaseMessageParam]:
        """Provider-agnostic list of messages."""
        return self._message_converter.from_provider(self.messages)

    @property
    def common_tools(self) -> list[Tool] | None:
        """Provider-agnostic tools."""
        from ...llm.tool import Tool

        if not self.tools:
            return None
        return [Tool(tool=tool) for tool in self.tools]  # pyright: ignore [reportAbstractUsage]

    @property
    def common_usage(self) -> Usage | None:
        """Provider-agnostic usage info."""
        return get_common_usage(
            self.input_tokens, self.cached_tokens, self.output_tokens
        )



"""Tests the `xai.call_response` module."""

from openai.types.chat import ChatCompletion
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.completion_usage import CompletionUsage

from mirascope.core.base.types import CostMetadata
from mirascope.core.xai.call_response import XAICallResponse


def test_xai_call_response_cost() -> None:
    choices = [
        Choice(
            finish_reason="stop",
            index=0,
            message=ChatCompletionMessage(content="content", role="assistant"),
        )
    ]
    usage = CompletionUsage(completion_tokens=1, prompt_tokens=1, total_tokens=2)
    completion = ChatCompletion(
        id="id",
        choices=choices,
        created=0,
        model="grok-3",
        object="chat.completion",
        usage=usage,
    )
    call_response = XAICallResponse(
        metadata={},
        response=completion,  # pyright: ignore [reportArgumentType]
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )

    assert call_response.cost == 1.4e-5
    assert call_response.cost_metadata == CostMetadata(input_tokens=1, output_tokens=1)



from typing import Any, cast

from mirascope.core.base import BaseCallResponseChunk
from mirascope.core.base.types import CostMetadata, FinishReason, Usage
from mirascope.llm.call_response_chunk import CallResponseChunk


class DummyCallResponseChunk(BaseCallResponseChunk[Any, str]):
    @property
    def content(self) -> str:
        return "chunk_content"

    @property
    def finish_reasons(self) -> list[str] | None:
        return ["stop"]

    @property
    def model(self) -> str | None: ...

    @property
    def id(self) -> str | None: ...

    @property
    def usage(self) -> Any: ...

    @property
    def input_tokens(self) -> int | float | None:
        return 2

    @property
    def cached_tokens(self) -> int | float | None:
        return 1

    @property
    def output_tokens(self) -> int | float | None:
        return 3

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        return cast(list[FinishReason], self.finish_reasons)

    @property
    def common_usage(self) -> Usage:
        input_tokens = int(self.input_tokens or 0)
        cached_tokens = int(self.cached_tokens or 0)
        output_tokens = int(self.output_tokens or 0)
        return Usage(
            input_tokens=input_tokens,
            cached_tokens=cached_tokens,
            output_tokens=output_tokens,
            total_tokens=input_tokens + output_tokens,
        )

    @property
    def cost_metadata(self) -> CostMetadata:
        return super().cost_metadata


def test_call_response_chunk():
    chunk_response_instance = DummyCallResponseChunk(chunk={})
    call_response_chunk_instance = CallResponseChunk(response=chunk_response_instance)  # pyright: ignore [reportAbstractUsage]
    assert call_response_chunk_instance.finish_reasons == ["stop"]
    assert call_response_chunk_instance.usage is not None
    assert call_response_chunk_instance.content == "chunk_content"
    assert call_response_chunk_instance.common_usage is not None
    assert call_response_chunk_instance.common_usage.total_tokens == 5
    assert call_response_chunk_instance.cost_metadata == CostMetadata(
        input_tokens=2,
        output_tokens=3,
        cached_tokens=1,
    )



"""The `XAIStream` class for convenience around streaming xAI LLM calls.

usage docs: learn/streams.md
"""

from collections.abc import AsyncGenerator, Generator

from ..openai import OpenAIStream
from .call_response import XAICallResponse
from .call_response_chunk import XAICallResponseChunk
from .tool import XAITool


class XAIStream(OpenAIStream):
    """A simple wrapper around `OpenAIStream`.

    Everything is the same except updates to the `construct_call_response` method and
    the `cost` property so that cost is properly calculated using xAI's cost
    calculation method. This ensures cost calculation works for non-OpenAI models.
    """

    _provider = "xai"

    def __iter__(
        self,
    ) -> Generator[tuple[XAICallResponseChunk, XAITool | None], None, None]:
        yield from super().__iter__()  # pyright: ignore [reportReturnType]

    def __aiter__(
        self,
    ) -> AsyncGenerator[tuple[XAICallResponseChunk, XAITool | None], None]:
        return super().__aiter__()  # pyright: ignore [reportReturnType] # pragma: no cover

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        response = self.construct_call_response()
        return response.cost

    def construct_call_response(self) -> XAICallResponse:
        openai_call_response = super().construct_call_response()
        response = XAICallResponse(
            metadata=openai_call_response.metadata,
            response=openai_call_response.response,
            tool_types=openai_call_response.tool_types,
            prompt_template=openai_call_response.prompt_template,
            fn_args=openai_call_response.fn_args,
            dynamic_config=openai_call_response.dynamic_config,
            messages=openai_call_response.messages,
            call_params=openai_call_response.call_params,
            call_kwargs=openai_call_response.call_kwargs,
            user_message_param=openai_call_response.user_message_param,
            start_time=openai_call_response.start_time,
            end_time=openai_call_response.end_time,
        )
        response._model = self.model
        return response



"""Calculate the cost of a completion using the Azure API."""

from ..base.types import CostMetadata


def calculate_cost(
    metadata: CostMetadata,
    model: str,
) -> float | None:
    """Calculate the cost of a completion using the Azure API."""
    return None



"""Protocols for reusable type hints."""

from collections.abc import (
    AsyncIterable,
    Callable,
    Iterable,
)
from enum import Enum
from typing import (
    Any,
    Literal,
    NoReturn,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from ..core import BaseDynamicConfig, BaseTool
from ..core.base import (
    BaseCallResponse,
    BaseCallResponseChunk,
    BaseType,
    CommonCallParams,
)
from ..core.base._utils._protocols import (
    AsyncLLMFunctionDecorator,
    LLMFunctionDecorator,
    SyncLLMFunctionDecorator,
)
from ..core.base.stream_config import StreamConfig
from ..core.base.types import LocalProvider, Provider
from .call_response import CallResponse
from .call_response_chunk import CallResponseChunk
from .stream import Stream

_P = ParamSpec("_P")
_R = TypeVar("_R", contravariant=True)

_BaseStreamT = TypeVar("_BaseStreamT", covariant=True)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType | Enum)
_SameSyncAndAsyncClientT = TypeVar("_SameSyncAndAsyncClientT", contravariant=True)
_SyncBaseClientT = TypeVar("_SyncBaseClientT", contravariant=True)
_AsyncBaseClientT = TypeVar("_AsyncBaseClientT", contravariant=True)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", contravariant=True)
_AsyncBaseDynamicConfigT = TypeVar("_AsyncBaseDynamicConfigT", contravariant=True)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", contravariant=True)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallResponseT = TypeVar(
    "_BaseCallResponseT", bound=BaseCallResponse, covariant=True
)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk, covariant=True
)


class _CallDecorator(
    Protocol[
        _BaseCallResponseT,
        _BaseCallResponseChunkT,
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallParamsT,
        _BaseStreamT,
        _SyncBaseClientT,
        _AsyncBaseClientT,
        _SameSyncAndAsyncClientT,
    ],
):
    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallResponseT,
        _BaseCallResponseT,
    ]: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _BaseCallResponseT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _BaseCallResponseT]: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _ParsedOutputT, _ParsedOutputT
    ]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseChunkT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _SyncBaseClientT
        | _AsyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _BaseStreamT, _BaseStreamT
    ]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _BaseStreamT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _BaseStreamT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseChunkT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _SyncBaseClientT
        | _AsyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _SyncBaseClientT
        | _AsyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _ResponseModelT, _ResponseModelT
    ]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _ResponseModelT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _ResponseModelT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _ParsedOutputT, _ParsedOutputT
    ]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        Iterable[_ResponseModelT],
        AsyncIterable[_ResponseModelT],
    ]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[
        _AsyncBaseDynamicConfigT, AsyncIterable[_ResponseModelT]
    ]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, Iterable[_ResponseModelT]]: ...

    @overload
    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT]
        | Callable[[_BaseCallResponseChunkT], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _AsyncBaseClientT
        | _SyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    def __call__(
        self,
        provider: Provider | LocalProvider,
        model: str,
        *,
        stream: bool | StreamConfig = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT] | None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT]
        | Callable[[_BaseCallResponseChunkT], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _AsyncBaseClientT
        | _SyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> (
        AsyncLLMFunctionDecorator[
            _AsyncBaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | AsyncIterable[_ResponseModelT],
        ]
        | SyncLLMFunctionDecorator[
            _BaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | Iterable[_ResponseModelT],
        ]
        | LLMFunctionDecorator[
            _BaseDynamicConfigT,
            _AsyncBaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | Iterable[_ResponseModelT],
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | AsyncIterable[_ResponseModelT],
        ]
    ): ...


CallDecorator: TypeAlias = _CallDecorator[
    CallResponse,
    CallResponseChunk,
    BaseDynamicConfig,
    BaseDynamicConfig,
    CommonCallParams,
    Stream,
    Any,
    Any,
    Any,
]



"""Tests the `anthropic.stream` module."""

import pytest
from anthropic.types import (
    Message,
    MessageDeltaUsage,
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
    TextBlock,
    TextDelta,
    ToolUseBlock,
    Usage,
)

try:
    from anthropic.types import (
        InputJsonDelta as InputJSONDelta,  # pyright: ignore [reportAttributeAccessIssue]
    )
except ImportError:
    from anthropic.types import (
        InputJSONDelta,  # pyright: ignore [reportAttributeAccessIssue]
    )

from anthropic.types.raw_message_delta_event import Delta

from mirascope.core.anthropic.call_params import AnthropicCallParams
from mirascope.core.anthropic.call_response import AnthropicCallResponse
from mirascope.core.anthropic.call_response_chunk import AnthropicCallResponseChunk
from mirascope.core.anthropic.stream import AnthropicStream
from mirascope.core.anthropic.tool import AnthropicTool


def test_anthropic_stream() -> None:
    """Tests the `AnthropicStream` class."""
    assert AnthropicStream._provider == "anthropic"

    class FormatBook(AnthropicTool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> None:
            """Dummy call."""

    chunks = [
        RawMessageStartEvent(
            message=Message(
                id="id",
                content=[],
                model="claude-3-5-sonnet-20240620",
                role="assistant",
                stop_reason=None,
                stop_sequence=None,
                type="message",
                usage=Usage(input_tokens=1, output_tokens=1),
            ),
            type="message_start",
        ),
        RawContentBlockStartEvent(
            content_block=ToolUseBlock(
                id="tool_id",
                input={},
                name="FormatBook",
                type="tool_use",
            ),
            index=0,
            type="content_block_start",
        ),
        RawContentBlockDeltaEvent(
            delta=InputJSONDelta(partial_json="", type="input_json_delta"),
            index=0,
            type="content_block_delta",
        ),
        RawContentBlockDeltaEvent(
            delta=InputJSONDelta(
                partial_json='{"'
                'title": "Sapiens: A Brief History of Humankind", '
                '"author": "Harari, Yuval Noah"}',
                type="input_json_delta",
            ),
            index=0,
            type="content_block_delta",
        ),
        RawContentBlockStopEvent(index=0, type="content_block_stop"),
        RawMessageDeltaEvent(
            delta=Delta(stop_reason="tool_use", stop_sequence=None),
            type="message_delta",
            usage=MessageDeltaUsage(output_tokens=1),
        ),
        RawMessageStopEvent(type="message_stop"),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = AnthropicCallResponseChunk(chunk=chunk)
            if isinstance(call_response_chunk.chunk, RawContentBlockStartEvent) and (
                tool_call := call_response_chunk.chunk.content_block
            ):
                tool_call = ToolUseBlock(
                    id="tool_id",
                    input={
                        "title": "Sapiens: A Brief History of Humankind",
                        "author": "Harari, Yuval Noah",
                    },
                    name="FormatBook",
                    type="tool_use",
                )
                yield (
                    call_response_chunk,
                    FormatBook.from_tool_call(tool_call),
                )
            else:
                yield call_response_chunk, None

    stream = AnthropicStream(
        stream=generator(),
        metadata={},
        tool_types=[FormatBook],
        call_response_type=AnthropicCallResponse,
        model="claude-3-5-sonnet-20240620",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params=AnthropicCallParams(max_tokens=1000),
        call_kwargs={},
    )

    with pytest.raises(
        ValueError, match="No stream response, check if the stream has been consumed."
    ):
        stream.construct_call_response()

    assert stream.cost is None
    for _ in stream:
        pass
    assert stream.cost == 3.3e-05

    format_book = FormatBook.from_tool_call(
        ToolUseBlock(
            id="tool_id",
            input={
                "title": "Sapiens: A Brief History of Humankind",
                "author": "Harari, Yuval Noah",
            },
            name="FormatBook",
            type="tool_use",
        )
    )
    assert format_book.tool_call is not None
    assert stream.message_param == {
        "role": "assistant",
        "content": [format_book.tool_call.model_dump()],
    }


def test_construct_call_response() -> None:
    class FormatBook(AnthropicTool):
        """Returns the title and author nicely formatted."""

        title: str
        author: str

        def call(self) -> None:
            """Dummy call."""

    chunks = [
        RawMessageStartEvent(
            message=Message(
                id="id",
                content=[],
                model="claude-3-5-sonnet-20240620",
                role="assistant",
                stop_reason="end_turn",
                stop_sequence=None,
                type="message",
                usage=Usage(input_tokens=1, output_tokens=1),
            ),
            type="message_start",
        ),
        RawContentBlockStartEvent(
            content_block=ToolUseBlock(
                id="tool_id",
                input={},
                name="FormatBook",
                type="tool_use",
            ),
            index=0,
            type="content_block_start",
        ),
        RawContentBlockDeltaEvent(
            delta=InputJSONDelta(partial_json="", type="input_json_delta"),
            index=0,
            type="content_block_delta",
        ),
        RawContentBlockDeltaEvent(
            delta=InputJSONDelta(
                partial_json='{"'
                'title": "Sapiens: A Brief History of Humankind", '
                '"author": "Harari, Yuval Noah"}',
                type="input_json_delta",
            ),
            index=0,
            type="content_block_delta",
        ),
        RawContentBlockStopEvent(index=0, type="content_block_stop"),
        RawContentBlockStartEvent(
            content_block=TextBlock(text="", type="text"),
            index=1,
            type="content_block_start",
        ),
        RawContentBlockDeltaEvent(
            delta=TextDelta(text="content", type="text_delta"),
            index=1,
            type="content_block_delta",
        ),
        RawContentBlockStopEvent(index=1, type="content_block_stop"),
        RawMessageDeltaEvent(
            delta=Delta(stop_reason="tool_use", stop_sequence=None),
            type="message_delta",
            usage=MessageDeltaUsage(output_tokens=1),
        ),
        RawMessageStopEvent(type="message_stop"),
    ]

    def generator():
        for chunk in chunks:
            call_response_chunk = AnthropicCallResponseChunk(chunk=chunk)
            if (
                isinstance(call_response_chunk.chunk, RawContentBlockStartEvent)
                and (tool_call := call_response_chunk.chunk.content_block)
                and tool_call.type == "tool_use"
            ):
                tool_call = ToolUseBlock(
                    id="tool_id",
                    input={
                        "title": "Sapiens: A Brief History of Humankind",
                        "author": "Harari, Yuval Noah",
                    },
                    name="FormatBook",
                    type="tool_use",
                )
                yield (
                    call_response_chunk,
                    FormatBook.from_tool_call(tool_call),
                )
            else:
                yield call_response_chunk, None

    stream = AnthropicStream(
        stream=generator(),
        metadata={},
        tool_types=[FormatBook],
        call_response_type=AnthropicCallResponse,
        model="claude-3-5-sonnet-20240620",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params=AnthropicCallParams(max_tokens=1000),
        call_kwargs={},
    )

    for _ in stream:
        pass

    tool_call = ToolUseBlock(
        id="tool_id",
        input={
            "title": "Sapiens: A Brief History of Humankind",
            "author": "Harari, Yuval Noah",
        },
        name="FormatBook",
        type="tool_use",
    )
    content = TextBlock(text="content", type="text")
    completion = Message(
        id="id",
        content=[content, tool_call],
        model="claude-3-5-sonnet-20240620",
        role="assistant",
        stop_reason="end_turn",
        stop_sequence=None,
        type="message",
        usage=Usage(input_tokens=1, output_tokens=2),
    )
    call_response = AnthropicCallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=AnthropicCallParams(max_tokens=1000),
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    constructed_call_response = stream.construct_call_response()
    assert constructed_call_response.response == call_response.response


def test_construct_call_response_string_content() -> None:
    """Tests the `construct_call_response` method handles string message_param.
    This should never occur since we convert all string message_param to TextBlock.
    """
    chunks = [
        RawMessageStartEvent(
            message=Message(
                id="id",
                content=[],
                model="claude-3-5-sonnet-20240620",
                role="assistant",
                stop_reason="end_turn",
                stop_sequence=None,
                type="message",
                usage=Usage(input_tokens=1, output_tokens=1),
            ),
            type="message_start",
        )
    ]

    def generator():  # pragma: no cover
        for chunk in chunks:
            call_response_chunk = AnthropicCallResponseChunk(chunk=chunk)
            yield call_response_chunk, None

    stream = AnthropicStream(
        stream=generator(),
        metadata={},
        tool_types=None,
        call_response_type=AnthropicCallResponse,
        model="claude-3-5-sonnet-20240620",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[{"role": "user", "content": "content"}],
        call_params=AnthropicCallParams(max_tokens=1000),
        call_kwargs={},
    )
    stream.message_param = {
        "role": "assistant",
        "content": "content",
    }
    constructed_call_response = stream.construct_call_response()
    assert constructed_call_response.message_param == {
        "role": "assistant",
        "content": [{"text": "content", "type": "text"}],
    }



import json
from typing import cast

from azure.ai.inference.models import (
    AssistantMessage,
    ChatRequestMessage,
    ContentItem,
    ImageContentItem,
    TextContentItem,
    ToolMessage,
    UserMessage,
)

from mirascope.core.azure._utils import convert_message_params
from mirascope.core.base import (
    BaseMessageParam,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)


def _parse_content(content: list[ContentItem]) -> list[TextPart | ImageURLPart]:
    converted_parts = []
    for part in content:
        if isinstance(part, TextContentItem):
            converted_parts.append(TextPart(type="text", text=part.text))
        elif isinstance(part, ImageContentItem):
            converted_parts.append(
                ImageURLPart(
                    type="image_url",
                    url=part.image_url.url,
                    detail=part.image_url.detail,
                )
            )

    return converted_parts


class AzureMessageParamConverter(BaseMessageParamConverter):
    """Converts between Azure `ChatRequestMessage` / `AssistantMessage` and Mirascope `BaseMessageParam`."""

    @staticmethod
    def to_provider(message_params: list[BaseMessageParam]) -> list[ChatRequestMessage]:
        """Convert from Mirascope `BaseMessageParam` to Azure's `ChatRequestMessage`."""
        return convert_message_params(
            cast(list[BaseMessageParam | ChatRequestMessage], message_params)
        )

    @staticmethod
    def from_provider(
        message_params: list[ChatRequestMessage],
    ) -> list[BaseMessageParam]:
        """
        Convert from Azure's `AssistantMessage` back to Mirascope `BaseMessageParam`.
        """
        converted: list[BaseMessageParam] = []
        for message_param in message_params:
            if isinstance(message_param, UserMessage):
                if isinstance(message_param.content, str):
                    converted.append(
                        BaseMessageParam(role="user", content=message_param.content)
                    )
                elif isinstance(message_param.content, list):
                    converted_parts = _parse_content(message_param.content)
                    converted.append(
                        BaseMessageParam(role="user", content=converted_parts)
                    )
            elif isinstance(message_param, AssistantMessage):
                converted_parts = []
                if tool_calls := message_param.tool_calls:
                    converted_parts.extend(
                        ToolCallPart(
                            type="tool_call",
                            name=tool_call.function.name,
                            id=tool_call.id,
                            args=json.loads(tool_call.function.arguments),
                        )
                        for tool_call in tool_calls
                    )
                if isinstance(message_param.content, str):
                    converted_parts.append(
                        TextPart(type="text", text=message_param.content)
                    )
                elif isinstance(message_param.content, list):
                    converted_parts = _parse_content(message_param.content)
                else:
                    converted_parts.append(TextPart(type="text", text=""))
                converted.append(
                    BaseMessageParam(
                        role="assistant",
                        content=converted_parts[0].text
                        if len(converted_parts) == 1
                        and isinstance(converted_parts[0], TextPart)
                        else converted_parts,
                    )
                )
            elif isinstance(message_param, ToolMessage):
                converted.append(
                    BaseMessageParam(
                        role="tool",
                        content=[
                            ToolResultPart(
                                type="tool_result",
                                name="",
                                content=message_param.content,
                                id=message_param.tool_call_id,
                                is_error=False,
                            )
                        ],
                    )
                )
        return converted



from typing import cast

from ...base.call_params import CommonCallParams
from ..call_params import MistralCallParams

MISTRAL_PARAM_MAPPING = {
    "temperature": "temperature",
    "max_tokens": "max_tokens",
    "top_p": "top_p",
    "seed": "random_seed",
    "stop": "stop",
}


def convert_common_call_params(common_params: CommonCallParams) -> MistralCallParams:
    """Convert CommonCallParams to Mistral parameters."""
    return cast(
        MistralCallParams,
        {
            MISTRAL_PARAM_MAPPING[key]: value
            for key, value in common_params.items()
            if key in MISTRAL_PARAM_MAPPING and value is not None
        },
    )



"""The Mirascope Core Functionality."""

from contextlib import suppress

from . import base, costs
from .base import (
    AudioPart,
    AudioURLPart,
    BaseCallResponse,
    BaseDynamicConfig,
    BaseMessageParam,
    BasePrompt,
    BaseStream,
    BaseTool,
    BaseToolKit,
    CacheControlPart,
    CostMetadata,
    DocumentPart,
    DocumentURLPart,
    FromCallArgs,
    ImagePart,
    ImageURLPart,
    LocalProvider,
    Messages,
    Provider,
    ResponseModelConfigDict,
    TextPart,
    ToolCallPart,
    ToolResultPart,
    merge_decorators,
    metadata,
    prompt_template,
    toolkit_tool,
)
from .costs import calculate_cost

with suppress(ImportError):
    from . import anthropic as anthropic

with suppress(ImportError):
    from . import cohere as cohere

with suppress(ImportError):
    from . import google as google

with suppress(ImportError):
    from . import gemini as gemini

with suppress(ImportError):
    from . import groq as groq

with suppress(ImportError):
    from . import litellm as litellm

with suppress(ImportError):
    from . import mistral as mistral

with suppress(ImportError):
    from . import openai as openai

with suppress(ImportError):
    from . import vertex as vertex

with suppress(ImportError):
    from . import azure as azure

__all__ = [
    "AudioPart",
    "AudioURLPart",
    "BaseCallResponse",
    "BaseCallResponse",
    "BaseDynamicConfig",
    "BaseMessageParam",
    "BasePrompt",
    "BaseStream",
    "BaseTool",
    "BaseToolKit",
    "CacheControlPart",
    "CostMetadata",
    "DocumentPart",
    "DocumentURLPart",
    "FromCallArgs",
    "ImagePart",
    "ImageURLPart",
    "LocalProvider",
    "Messages",
    "Provider",
    "ResponseModelConfigDict",
    "TextPart",
    "ToolCallPart",
    "ToolResultPart",
    "anthropic",
    "azure",
    "base",
    "calculate_cost",
    "cohere",
    "costs",
    "gemini",
    "google",
    "groq",
    "litellm",
    "merge_decorators",
    "metadata",
    "mistral",
    "openai",
    "prompt_template",
    "toolkit_tool",
    "vertex",
]



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from azure.core.credentials import AzureKeyCredential

if TYPE_CHECKING:
    from azure.identity import (  # pyright: ignore [reportMissingImports]
        DefaultAzureCredential,
    )
else:
    AzureDefaultCredential = None


def _get_key_from_env() -> str | None:
    return os.environ.get("AZURE_INFERENCE_CREDENTIAL")


def _get_azure_default_credential() -> DefaultAzureCredential:
    try:
        from azure.identity import DefaultAzureCredential
    except ImportError:
        raise ImportError(
            "To use the AzureDefaultCredential, please install the azure-identity package with `pip install azure-identity`."
        )
    return DefaultAzureCredential()


def get_credential() -> DefaultAzureCredential | AzureKeyCredential:
    if (credential := _get_key_from_env()) is None:
        return _get_azure_default_credential()
    return AzureKeyCredential(credential)



"""Calculate the cost of a Gemini API call."""

from ..base.types import CostMetadata


def calculate_cost(
    metadata: CostMetadata,
    model: str,
) -> float | None:
    """Calculate the cost of a Gemini API call.

    https://ai.google.dev/pricing#1_5flash

    Model                 Input (<=128K)     Output (<=128K)    Input (>128K)      Output (>128K)
    gemini-1.5-flash      $0.075 / 1M        $0.3 / 1M          $0.15 / 1M         $0.6 / 1M
    gemini-1.5-flash-8b   $0.0375 / 1M       $0.15 / 1M         $0.075 / 1M        $0.3 / 1M
    gemini-1.5-pro        $1.25 / 1M         $5.0 / 1M          $2.5 / 1M          $10.0 / 1M
    gemini-1.0-pro        $0.50 / 1M         $1.5 / 1M          $0.5 / 1M          $1.5 / 1M
    """
    pricing = {
        "gemini-1.5-flash": {
            "prompt_short": 0.000_000_075,
            "completion_short": 0.000_000_3,
            "prompt_long": 0.000_000_15,
            "completion_long": 0.000_000_6,
        },
        "gemini-1.5-flash-8b": {
            "prompt_short": 0.000_000_037_5,
            "completion_short": 0.000_000_15,
            "prompt_long": 0.000_000_075,
            "completion_long": 0.000_000_3,
        },
        "gemini-1.5-pro": {
            "prompt_short": 0.000_001_25,
            "completion_short": 0.000_005,
            "prompt_long": 0.000_002_5,
            "completion_long": 0.000_01,
        },
        "gemini-1.0-pro": {
            "prompt_short": 0.000_000_5,
            "completion_short": 0.000_001_5,
            "prompt_long": 0.000_000_5,
            "completion_long": 0.000_001_5,
        },
    }

    if metadata.input_tokens is None or metadata.output_tokens is None:
        return None

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    # Determine if we're using long context pricing
    use_long_context = metadata.input_tokens > 128_000

    prompt_price = model_pricing["prompt_long" if use_long_context else "prompt_short"]
    completion_price = model_pricing[
        "completion_long" if use_long_context else "completion_short"
    ]

    prompt_cost = metadata.input_tokens * prompt_price
    completion_cost = metadata.output_tokens * completion_price
    total_cost = prompt_cost + completion_cost

    return total_cost



from ._utils import Context
from .realtime import Realtime
from .recording import async_input, record, record_as_stream
from .tool import OpenAIRealtimeTool

__all__ = [
    "Context",
    "OpenAIRealtimeTool",
    "Realtime",
    "async_input",
    "record",
    "record_as_stream",
]



"""Mirascope Base Classes."""

from . import _partial, _utils
from ._call_factory import call_factory
from ._utils import BaseType
from .call_kwargs import BaseCallKwargs
from .call_params import BaseCallParams, CommonCallParams
from .call_response import BaseCallResponse, transform_tool_outputs
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .from_call_args import FromCallArgs
from .merge_decorators import merge_decorators
from .message_param import (
    AudioPart,
    AudioURLPart,
    BaseMessageParam,
    CacheControlPart,
    DocumentPart,
    DocumentURLPart,
    ImagePart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from .messages import Messages
from .metadata import Metadata
from .prompt import BasePrompt, metadata, prompt_template
from .response_model_config_dict import ResponseModelConfigDict
from .stream import BaseStream
from .structured_stream import BaseStructuredStream
from .tool import BaseTool, GenerateJsonSchemaNoTitles, ToolConfig
from .toolkit import BaseToolKit, toolkit_tool
from .types import (
    AudioSegment,
    CostMetadata,
    JsonableType,
    LocalProvider,
    Provider,
    Usage,
)

__all__ = [
    "AudioPart",
    "AudioSegment",
    "AudioURLPart",
    "BaseCallKwargs",
    "BaseCallParams",
    "BaseCallResponse",
    "BaseCallResponseChunk",
    "BaseDynamicConfig",
    "BaseMessageParam",
    "BasePrompt",
    "BaseStream",
    "BaseStructuredStream",
    "BaseTool",
    "BaseToolKit",
    "BaseType",
    "CacheControlPart",
    "CommonCallParams",
    "CostMetadata",
    "DocumentPart",
    "DocumentURLPart",
    "FromCallArgs",
    "GenerateJsonSchemaNoTitles",
    "ImagePart",
    "ImageURLPart",
    "JsonableType",
    "LocalProvider",
    "Messages",
    "Metadata",
    "ResponseModelConfigDict",
    "TextPart",
    "ToolCallPart",
    "ToolConfig",
    "ToolResultPart",
    "Usage",
    "_partial",
    "_utils",
    "call_factory",
    "merge_decorators",
    "metadata",
    "prompt_template",
    "toolkit_tool",
    "transform_tool_outputs",
]
